import sys
from datetime import datetime
from typing import Dict, List, Optional

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

from localstack.aws.api import RequestContext, ServiceException, ServiceRequest, handler

ARN = str
BackendRole = str
Boolean = bool
CloudWatchLogsLogGroupArn = str
CommitMessage = str
ConnectionAlias = str
CrossClusterSearchConnectionId = str
CrossClusterSearchConnectionStatusMessage = str
DeploymentType = str
DescribePackagesFilterValue = str
DomainId = str
DomainName = str
DomainNameFqdn = str
Double = float
DryRun = bool
ElasticsearchVersionString = str
ErrorMessage = str
ErrorType = str
GUID = str
IdentityPoolId = str
InstanceCount = int
InstanceRole = str
Integer = int
IntegerClass = int
Issue = str
KmsKeyId = str
LimitName = str
LimitValue = str
MaxResults = int
MaximumInstanceCount = int
Message = str
MinimumInstanceCount = int
NextToken = str
NonEmptyString = str
OwnerId = str
PackageDescription = str
PackageID = str
PackageName = str
PackageVersion = str
Password = str
PolicyDocument = str
ReferencePath = str
Region = str
ReservationToken = str
RoleArn = str
S3BucketName = str
S3Key = str
SAMLEntityId = str
SAMLMetadata = str
ScheduledAutoTuneDescription = str
ServiceUrl = str
StorageSubTypeName = str
StorageTypeName = str
String = str
TagKey = str
TagValue = str
UIntValue = int
UpgradeName = str
UserPoolId = str
Username = str


class AutoTuneDesiredState(str):
    ENABLED = "ENABLED"
    DISABLED = "DISABLED"


class AutoTuneState(str):
    ENABLED = "ENABLED"
    DISABLED = "DISABLED"
    ENABLE_IN_PROGRESS = "ENABLE_IN_PROGRESS"
    DISABLE_IN_PROGRESS = "DISABLE_IN_PROGRESS"
    DISABLED_AND_ROLLBACK_SCHEDULED = "DISABLED_AND_ROLLBACK_SCHEDULED"
    DISABLED_AND_ROLLBACK_IN_PROGRESS = "DISABLED_AND_ROLLBACK_IN_PROGRESS"
    DISABLED_AND_ROLLBACK_COMPLETE = "DISABLED_AND_ROLLBACK_COMPLETE"
    DISABLED_AND_ROLLBACK_ERROR = "DISABLED_AND_ROLLBACK_ERROR"
    ERROR = "ERROR"


class AutoTuneType(str):
    SCHEDULED_ACTION = "SCHEDULED_ACTION"


class DeploymentStatus(str):
    PENDING_UPDATE = "PENDING_UPDATE"
    IN_PROGRESS = "IN_PROGRESS"
    COMPLETED = "COMPLETED"
    NOT_ELIGIBLE = "NOT_ELIGIBLE"
    ELIGIBLE = "ELIGIBLE"


class DescribePackagesFilterName(str):
    PackageID = "PackageID"
    PackageName = "PackageName"
    PackageStatus = "PackageStatus"


class DomainPackageStatus(str):
    ASSOCIATING = "ASSOCIATING"
    ASSOCIATION_FAILED = "ASSOCIATION_FAILED"
    ACTIVE = "ACTIVE"
    DISSOCIATING = "DISSOCIATING"
    DISSOCIATION_FAILED = "DISSOCIATION_FAILED"


class ESPartitionInstanceType(str):
    m3_medium_elasticsearch = "m3.medium.elasticsearch"
    m3_large_elasticsearch = "m3.large.elasticsearch"
    m3_xlarge_elasticsearch = "m3.xlarge.elasticsearch"
    m3_2xlarge_elasticsearch = "m3.2xlarge.elasticsearch"
    m4_large_elasticsearch = "m4.large.elasticsearch"
    m4_xlarge_elasticsearch = "m4.xlarge.elasticsearch"
    m4_2xlarge_elasticsearch = "m4.2xlarge.elasticsearch"
    m4_4xlarge_elasticsearch = "m4.4xlarge.elasticsearch"
    m4_10xlarge_elasticsearch = "m4.10xlarge.elasticsearch"
    m5_large_elasticsearch = "m5.large.elasticsearch"
    m5_xlarge_elasticsearch = "m5.xlarge.elasticsearch"
    m5_2xlarge_elasticsearch = "m5.2xlarge.elasticsearch"
    m5_4xlarge_elasticsearch = "m5.4xlarge.elasticsearch"
    m5_12xlarge_elasticsearch = "m5.12xlarge.elasticsearch"
    r5_large_elasticsearch = "r5.large.elasticsearch"
    r5_xlarge_elasticsearch = "r5.xlarge.elasticsearch"
    r5_2xlarge_elasticsearch = "r5.2xlarge.elasticsearch"
    r5_4xlarge_elasticsearch = "r5.4xlarge.elasticsearch"
    r5_12xlarge_elasticsearch = "r5.12xlarge.elasticsearch"
    c5_large_elasticsearch = "c5.large.elasticsearch"
    c5_xlarge_elasticsearch = "c5.xlarge.elasticsearch"
    c5_2xlarge_elasticsearch = "c5.2xlarge.elasticsearch"
    c5_4xlarge_elasticsearch = "c5.4xlarge.elasticsearch"
    c5_9xlarge_elasticsearch = "c5.9xlarge.elasticsearch"
    c5_18xlarge_elasticsearch = "c5.18xlarge.elasticsearch"
    ultrawarm1_medium_elasticsearch = "ultrawarm1.medium.elasticsearch"
    ultrawarm1_large_elasticsearch = "ultrawarm1.large.elasticsearch"
    t2_micro_elasticsearch = "t2.micro.elasticsearch"
    t2_small_elasticsearch = "t2.small.elasticsearch"
    t2_medium_elasticsearch = "t2.medium.elasticsearch"
    r3_large_elasticsearch = "r3.large.elasticsearch"
    r3_xlarge_elasticsearch = "r3.xlarge.elasticsearch"
    r3_2xlarge_elasticsearch = "r3.2xlarge.elasticsearch"
    r3_4xlarge_elasticsearch = "r3.4xlarge.elasticsearch"
    r3_8xlarge_elasticsearch = "r3.8xlarge.elasticsearch"
    i2_xlarge_elasticsearch = "i2.xlarge.elasticsearch"
    i2_2xlarge_elasticsearch = "i2.2xlarge.elasticsearch"
    d2_xlarge_elasticsearch = "d2.xlarge.elasticsearch"
    d2_2xlarge_elasticsearch = "d2.2xlarge.elasticsearch"
    d2_4xlarge_elasticsearch = "d2.4xlarge.elasticsearch"
    d2_8xlarge_elasticsearch = "d2.8xlarge.elasticsearch"
    c4_large_elasticsearch = "c4.large.elasticsearch"
    c4_xlarge_elasticsearch = "c4.xlarge.elasticsearch"
    c4_2xlarge_elasticsearch = "c4.2xlarge.elasticsearch"
    c4_4xlarge_elasticsearch = "c4.4xlarge.elasticsearch"
    c4_8xlarge_elasticsearch = "c4.8xlarge.elasticsearch"
    r4_large_elasticsearch = "r4.large.elasticsearch"
    r4_xlarge_elasticsearch = "r4.xlarge.elasticsearch"
    r4_2xlarge_elasticsearch = "r4.2xlarge.elasticsearch"
    r4_4xlarge_elasticsearch = "r4.4xlarge.elasticsearch"
    r4_8xlarge_elasticsearch = "r4.8xlarge.elasticsearch"
    r4_16xlarge_elasticsearch = "r4.16xlarge.elasticsearch"
    i3_large_elasticsearch = "i3.large.elasticsearch"
    i3_xlarge_elasticsearch = "i3.xlarge.elasticsearch"
    i3_2xlarge_elasticsearch = "i3.2xlarge.elasticsearch"
    i3_4xlarge_elasticsearch = "i3.4xlarge.elasticsearch"
    i3_8xlarge_elasticsearch = "i3.8xlarge.elasticsearch"
    i3_16xlarge_elasticsearch = "i3.16xlarge.elasticsearch"


class ESWarmPartitionInstanceType(str):
    ultrawarm1_medium_elasticsearch = "ultrawarm1.medium.elasticsearch"
    ultrawarm1_large_elasticsearch = "ultrawarm1.large.elasticsearch"


class EngineType(str):
    OpenSearch = "OpenSearch"
    Elasticsearch = "Elasticsearch"


class InboundCrossClusterSearchConnectionStatusCode(str):
    PENDING_ACCEPTANCE = "PENDING_ACCEPTANCE"
    APPROVED = "APPROVED"
    REJECTING = "REJECTING"
    REJECTED = "REJECTED"
    DELETING = "DELETING"
    DELETED = "DELETED"


class LogType(str):
    INDEX_SLOW_LOGS = "INDEX_SLOW_LOGS"
    SEARCH_SLOW_LOGS = "SEARCH_SLOW_LOGS"
    ES_APPLICATION_LOGS = "ES_APPLICATION_LOGS"
    AUDIT_LOGS = "AUDIT_LOGS"


class OptionState(str):
    RequiresIndexDocuments = "RequiresIndexDocuments"
    Processing = "Processing"
    Active = "Active"


class OutboundCrossClusterSearchConnectionStatusCode(str):
    PENDING_ACCEPTANCE = "PENDING_ACCEPTANCE"
    VALIDATING = "VALIDATING"
    VALIDATION_FAILED = "VALIDATION_FAILED"
    PROVISIONING = "PROVISIONING"
    ACTIVE = "ACTIVE"
    REJECTED = "REJECTED"
    DELETING = "DELETING"
    DELETED = "DELETED"


class PackageStatus(str):
    COPYING = "COPYING"
    COPY_FAILED = "COPY_FAILED"
    VALIDATING = "VALIDATING"
    VALIDATION_FAILED = "VALIDATION_FAILED"
    AVAILABLE = "AVAILABLE"
    DELETING = "DELETING"
    DELETED = "DELETED"
    DELETE_FAILED = "DELETE_FAILED"


class PackageType(str):
    TXT_DICTIONARY = "TXT-DICTIONARY"


class ReservedElasticsearchInstancePaymentOption(str):
    ALL_UPFRONT = "ALL_UPFRONT"
    PARTIAL_UPFRONT = "PARTIAL_UPFRONT"
    NO_UPFRONT = "NO_UPFRONT"


class RollbackOnDisable(str):
    NO_ROLLBACK = "NO_ROLLBACK"
    DEFAULT_ROLLBACK = "DEFAULT_ROLLBACK"


class ScheduledAutoTuneActionType(str):
    JVM_HEAP_SIZE_TUNING = "JVM_HEAP_SIZE_TUNING"
    JVM_YOUNG_GEN_TUNING = "JVM_YOUNG_GEN_TUNING"


class ScheduledAutoTuneSeverityType(str):
    LOW = "LOW"
    MEDIUM = "MEDIUM"
    HIGH = "HIGH"


class TLSSecurityPolicy(str):
    Policy_Min_TLS_1_0_2019_07 = "Policy-Min-TLS-1-0-2019-07"
    Policy_Min_TLS_1_2_2019_07 = "Policy-Min-TLS-1-2-2019-07"


class TimeUnit(str):
    HOURS = "HOURS"


class UpgradeStatus(str):
    IN_PROGRESS = "IN_PROGRESS"
    SUCCEEDED = "SUCCEEDED"
    SUCCEEDED_WITH_ISSUES = "SUCCEEDED_WITH_ISSUES"
    FAILED = "FAILED"


class UpgradeStep(str):
    PRE_UPGRADE_CHECK = "PRE_UPGRADE_CHECK"
    SNAPSHOT = "SNAPSHOT"
    UPGRADE = "UPGRADE"


class VolumeType(str):
    standard = "standard"
    gp2 = "gp2"
    io1 = "io1"


class AccessDeniedException(ServiceException):
    """An error occurred because user does not have permissions to access the
    resource. Returns HTTP status code 403.
    """

    pass


class BaseException(ServiceException):
    """An error occurred while processing the request."""

    message: Optional[ErrorMessage]


class ConflictException(ServiceException):
    """An error occurred because the client attempts to remove a resource that
    is currently in use. Returns HTTP status code 409.
    """

    pass


class DisabledOperationException(ServiceException):
    """An error occured because the client wanted to access a not supported
    operation. Gives http status code of 409.
    """

    pass


class InternalException(ServiceException):
    """The request processing has failed because of an unknown error, exception
    or failure (the failure is internal to the service) . Gives http status
    code of 500.
    """

    pass


class InvalidPaginationTokenException(ServiceException):
    """The request processing has failed because of invalid pagination token
    provided by customer. Returns an HTTP status code of 400.
    """

    pass


class InvalidTypeException(ServiceException):
    """An exception for trying to create or access sub-resource that is either
    invalid or not supported. Gives http status code of 409.
    """

    pass


class LimitExceededException(ServiceException):
    """An exception for trying to create more than allowed resources or
    sub-resources. Gives http status code of 409.
    """

    pass


class ResourceAlreadyExistsException(ServiceException):
    """An exception for creating a resource that already exists. Gives http
    status code of 400.
    """

    pass


class ResourceNotFoundException(ServiceException):
    """An exception for accessing or deleting a resource that does not exist.
    Gives http status code of 400.
    """

    pass


class ValidationException(ServiceException):
    """An exception for missing / invalid input fields. Gives http status code
    of 400.
    """

    pass


class AcceptInboundCrossClusterSearchConnectionRequest(ServiceRequest):
    """Container for the parameters to the
    ``AcceptInboundCrossClusterSearchConnection`` operation.
    """

    CrossClusterSearchConnectionId: CrossClusterSearchConnectionId


class InboundCrossClusterSearchConnectionStatus(TypedDict, total=False):
    """Specifies the coonection status of an inbound cross-cluster search
    connection.
    """

    StatusCode: Optional[InboundCrossClusterSearchConnectionStatusCode]
    Message: Optional[CrossClusterSearchConnectionStatusMessage]


class DomainInformation(TypedDict, total=False):
    OwnerId: Optional[OwnerId]
    DomainName: DomainName
    Region: Optional[Region]


class InboundCrossClusterSearchConnection(TypedDict, total=False):
    """Specifies details of an inbound connection."""

    SourceDomainInfo: Optional[DomainInformation]
    DestinationDomainInfo: Optional[DomainInformation]
    CrossClusterSearchConnectionId: Optional[CrossClusterSearchConnectionId]
    ConnectionStatus: Optional[InboundCrossClusterSearchConnectionStatus]


class AcceptInboundCrossClusterSearchConnectionResponse(TypedDict, total=False):
    """The result of a ``AcceptInboundCrossClusterSearchConnection`` operation.
    Contains details of accepted inbound connection.
    """

    CrossClusterSearchConnection: Optional[InboundCrossClusterSearchConnection]


UpdateTimestamp = datetime


class OptionStatus(TypedDict, total=False):
    """Provides the current status of the entity."""

    CreationDate: UpdateTimestamp
    UpdateDate: UpdateTimestamp
    UpdateVersion: Optional[UIntValue]
    State: OptionState
    PendingDeletion: Optional[Boolean]


class AccessPoliciesStatus(TypedDict, total=False):
    """The configured access rules for the domain's document and search
    endpoints, and the current status of those rules.
    """

    Options: PolicyDocument
    Status: OptionStatus


class Tag(TypedDict, total=False):
    """Specifies a key value pair for a resource tag."""

    Key: TagKey
    Value: TagValue


TagList = List[Tag]


class AddTagsRequest(ServiceRequest):
    """Container for the parameters to the ``AddTags`` operation. Specify the
    tags that you want to attach to the Elasticsearch domain.
    """

    ARN: ARN
    TagList: TagList


LimitValueList = List[LimitValue]


class AdditionalLimit(TypedDict, total=False):
    """List of limits that are specific to a given InstanceType and for each of
    it's ``InstanceRole`` .
    """

    LimitName: Optional[LimitName]
    LimitValues: Optional[LimitValueList]


AdditionalLimitList = List[AdditionalLimit]
AdvancedOptions = Dict[String, String]


class AdvancedOptionsStatus(TypedDict, total=False):
    """Status of the advanced options for the specified Elasticsearch domain.
    Currently, the following advanced options are available:

    -  Option to allow references to indices in an HTTP request body. Must
       be ``false`` when configuring access to individual sub-resources. By
       default, the value is ``true``. See `Configuration Advanced
       Options <http://docs.aws.amazon.com/elasticsearch-service/latest/developerguide/es-createupdatedomains.html#es-createdomain-configure-advanced-options>`__
       for more information.
    -  Option to specify the percentage of heap space that is allocated to
       field data. By default, this setting is unbounded.

    For more information, see `Configuring Advanced
    Options <http://docs.aws.amazon.com/elasticsearch-service/latest/developerguide/es-createupdatedomains.html#es-createdomain-configure-advanced-options>`__.
    """

    Options: AdvancedOptions
    Status: OptionStatus


DisableTimestamp = datetime


class SAMLIdp(TypedDict, total=False):
    """Specifies the SAML Identity Provider's information."""

    MetadataContent: SAMLMetadata
    EntityId: SAMLEntityId


class SAMLOptionsOutput(TypedDict, total=False):
    """Describes the SAML application configured for the domain."""

    Enabled: Optional[Boolean]
    Idp: Optional[SAMLIdp]
    SubjectKey: Optional[String]
    RolesKey: Optional[String]
    SessionTimeoutMinutes: Optional[IntegerClass]


class AdvancedSecurityOptions(TypedDict, total=False):
    """Specifies the advanced security configuration: whether advanced security
    is enabled, whether the internal database option is enabled.
    """

    Enabled: Optional[Boolean]
    InternalUserDatabaseEnabled: Optional[Boolean]
    SAMLOptions: Optional[SAMLOptionsOutput]
    AnonymousAuthDisableDate: Optional[DisableTimestamp]
    AnonymousAuthEnabled: Optional[Boolean]


class SAMLOptionsInput(TypedDict, total=False):
    """Specifies the SAML application configuration for the domain."""

    Enabled: Optional[Boolean]
    Idp: Optional[SAMLIdp]
    MasterUserName: Optional[Username]
    MasterBackendRole: Optional[BackendRole]
    SubjectKey: Optional[String]
    RolesKey: Optional[String]
    SessionTimeoutMinutes: Optional[IntegerClass]


class MasterUserOptions(TypedDict, total=False):
    """Credentials for the master user: username and password, ARN, or both."""

    MasterUserARN: Optional[ARN]
    MasterUserName: Optional[Username]
    MasterUserPassword: Optional[Password]


class AdvancedSecurityOptionsInput(TypedDict, total=False):
    """Specifies the advanced security configuration: whether advanced security
    is enabled, whether the internal database option is enabled, master
    username and password (if internal database is enabled), and master user
    ARN (if IAM is enabled).
    """

    Enabled: Optional[Boolean]
    InternalUserDatabaseEnabled: Optional[Boolean]
    MasterUserOptions: Optional[MasterUserOptions]
    SAMLOptions: Optional[SAMLOptionsInput]
    AnonymousAuthEnabled: Optional[Boolean]


class AdvancedSecurityOptionsStatus(TypedDict, total=False):
    """Specifies the status of advanced security options for the specified
    Elasticsearch domain.
    """

    Options: AdvancedSecurityOptions
    Status: OptionStatus


class AssociatePackageRequest(ServiceRequest):
    """Container for request parameters to ``AssociatePackage`` operation."""

    PackageID: PackageID
    DomainName: DomainName


class ErrorDetails(TypedDict, total=False):
    ErrorType: Optional[ErrorType]
    ErrorMessage: Optional[ErrorMessage]


LastUpdated = datetime


class DomainPackageDetails(TypedDict, total=False):
    """Information on a package that is associated with a domain."""

    PackageID: Optional[PackageID]
    PackageName: Optional[PackageName]
    PackageType: Optional[PackageType]
    LastUpdated: Optional[LastUpdated]
    DomainName: Optional[DomainName]
    DomainPackageStatus: Optional[DomainPackageStatus]
    PackageVersion: Optional[PackageVersion]
    ReferencePath: Optional[ReferencePath]
    ErrorDetails: Optional[ErrorDetails]


class AssociatePackageResponse(TypedDict, total=False):
    """Container for response returned by ``AssociatePackage`` operation."""

    DomainPackageDetails: Optional[DomainPackageDetails]


AutoTuneDate = datetime


class ScheduledAutoTuneDetails(TypedDict, total=False):
    """Specifies details of the scheduled Auto-Tune action. See the `Developer
    Guide <https://docs.aws.amazon.com/elasticsearch-service/latest/developerguide/auto-tune.html>`__
    for more information.
    """

    Date: Optional[AutoTuneDate]
    ActionType: Optional[ScheduledAutoTuneActionType]
    Action: Optional[ScheduledAutoTuneDescription]
    Severity: Optional[ScheduledAutoTuneSeverityType]


class AutoTuneDetails(TypedDict, total=False):
    """Specifies details of the Auto-Tune action. See the `Developer
    Guide <https://docs.aws.amazon.com/elasticsearch-service/latest/developerguide/auto-tune.html>`__
    for more information.
    """

    ScheduledAutoTuneDetails: Optional[ScheduledAutoTuneDetails]


class AutoTune(TypedDict, total=False):
    """Specifies Auto-Tune type and Auto-Tune action details."""

    AutoTuneType: Optional[AutoTuneType]
    AutoTuneDetails: Optional[AutoTuneDetails]


AutoTuneList = List[AutoTune]
DurationValue = int


class Duration(TypedDict, total=False):
    """Specifies maintenance schedule duration: duration value and duration
    unit. See the `Developer
    Guide <https://docs.aws.amazon.com/elasticsearch-service/latest/developerguide/auto-tune.html>`__
    for more information.
    """

    Value: Optional[DurationValue]
    Unit: Optional[TimeUnit]


StartAt = datetime


class AutoTuneMaintenanceSchedule(TypedDict, total=False):
    """Specifies Auto-Tune maitenance schedule. See the `Developer
    Guide <https://docs.aws.amazon.com/elasticsearch-service/latest/developerguide/auto-tune.html>`__
    for more information.
    """

    StartAt: Optional[StartAt]
    Duration: Optional[Duration]
    CronExpressionForRecurrence: Optional[String]


AutoTuneMaintenanceScheduleList = List[AutoTuneMaintenanceSchedule]


class AutoTuneOptions(TypedDict, total=False):
    """Specifies the Auto-Tune options: the Auto-Tune desired state for the
    domain, rollback state when disabling Auto-Tune options and list of
    maintenance schedules.
    """

    DesiredState: Optional[AutoTuneDesiredState]
    RollbackOnDisable: Optional[RollbackOnDisable]
    MaintenanceSchedules: Optional[AutoTuneMaintenanceScheduleList]


class AutoTuneOptionsInput(TypedDict, total=False):
    """Specifies the Auto-Tune options: the Auto-Tune desired state for the
    domain and list of maintenance schedules.
    """

    DesiredState: Optional[AutoTuneDesiredState]
    MaintenanceSchedules: Optional[AutoTuneMaintenanceScheduleList]


class AutoTuneOptionsOutput(TypedDict, total=False):
    """Specifies the Auto-Tune options: the Auto-Tune desired state for the
    domain and list of maintenance schedules.
    """

    State: Optional[AutoTuneState]
    ErrorMessage: Optional[String]


class AutoTuneStatus(TypedDict, total=False):
    """Provides the current status of the Auto-Tune options."""

    CreationDate: UpdateTimestamp
    UpdateDate: UpdateTimestamp
    UpdateVersion: Optional[UIntValue]
    State: AutoTuneState
    ErrorMessage: Optional[String]
    PendingDeletion: Optional[Boolean]


class AutoTuneOptionsStatus(TypedDict, total=False):
    """Specifies the status of Auto-Tune options for the specified
    Elasticsearch domain.
    """

    Options: Optional[AutoTuneOptions]
    Status: Optional[AutoTuneStatus]


class CancelElasticsearchServiceSoftwareUpdateRequest(ServiceRequest):
    """Container for the parameters to the
    ``CancelElasticsearchServiceSoftwareUpdate`` operation. Specifies the
    name of the Elasticsearch domain that you wish to cancel a service
    software update on.
    """

    DomainName: DomainName


DeploymentCloseDateTimeStamp = datetime


class ServiceSoftwareOptions(TypedDict, total=False):
    """The current options of an Elasticsearch domain service software options."""

    CurrentVersion: Optional[String]
    NewVersion: Optional[String]
    UpdateAvailable: Optional[Boolean]
    Cancellable: Optional[Boolean]
    UpdateStatus: Optional[DeploymentStatus]
    Description: Optional[String]
    AutomatedUpdateDate: Optional[DeploymentCloseDateTimeStamp]
    OptionalDeployment: Optional[Boolean]


class CancelElasticsearchServiceSoftwareUpdateResponse(TypedDict, total=False):
    """The result of a ``CancelElasticsearchServiceSoftwareUpdate`` operation.
    Contains the status of the update.
    """

    ServiceSoftwareOptions: Optional[ServiceSoftwareOptions]


class CognitoOptions(TypedDict, total=False):
    """Options to specify the Cognito user and identity pools for Kibana
    authentication. For more information, see `Amazon Cognito Authentication
    for
    Kibana <http://docs.aws.amazon.com/elasticsearch-service/latest/developerguide/es-cognito-auth.html>`__.
    """

    Enabled: Optional[Boolean]
    UserPoolId: Optional[UserPoolId]
    IdentityPoolId: Optional[IdentityPoolId]
    RoleArn: Optional[RoleArn]


class CognitoOptionsStatus(TypedDict, total=False):
    """Status of the Cognito options for the specified Elasticsearch domain."""

    Options: CognitoOptions
    Status: OptionStatus


class ColdStorageOptions(TypedDict, total=False):
    """Specifies the configuration for cold storage options such as enabled"""

    Enabled: Boolean


ElasticsearchVersionList = List[ElasticsearchVersionString]


class CompatibleVersionsMap(TypedDict, total=False):
    """A map from an ``ElasticsearchVersion`` to a list of compatible
    ``ElasticsearchVersion`` s to which the domain can be upgraded.
    """

    SourceVersion: Optional[ElasticsearchVersionString]
    TargetVersions: Optional[ElasticsearchVersionList]


CompatibleElasticsearchVersionsList = List[CompatibleVersionsMap]


class DomainEndpointOptions(TypedDict, total=False):
    """Options to configure endpoint for the Elasticsearch domain."""

    EnforceHTTPS: Optional[Boolean]
    TLSSecurityPolicy: Optional[TLSSecurityPolicy]
    CustomEndpointEnabled: Optional[Boolean]
    CustomEndpoint: Optional[DomainNameFqdn]
    CustomEndpointCertificateArn: Optional[ARN]


class LogPublishingOption(TypedDict, total=False):
    """| Log Publishing option that is set for given domain.
    | Attributes and their details:

    -  CloudWatchLogsLogGroupArn: ARN of the Cloudwatch log group to which
       log needs to be published.
    -  Enabled: Whether the log publishing for given log type is enabled or
       not
    """

    CloudWatchLogsLogGroupArn: Optional[CloudWatchLogsLogGroupArn]
    Enabled: Optional[Boolean]


LogPublishingOptions = Dict[LogType, LogPublishingOption]


class NodeToNodeEncryptionOptions(TypedDict, total=False):
    """Specifies the node-to-node encryption options."""

    Enabled: Optional[Boolean]


class EncryptionAtRestOptions(TypedDict, total=False):
    """Specifies the Encryption At Rest Options."""

    Enabled: Optional[Boolean]
    KmsKeyId: Optional[KmsKeyId]


StringList = List[String]


class VPCOptions(TypedDict, total=False):
    """Options to specify the subnets and security groups for VPC endpoint. For
    more information, see `VPC Endpoints for Amazon Elasticsearch Service
    Domains <http://docs.aws.amazon.com/elasticsearch-service/latest/developerguide/es-vpc.html>`__.
    """

    SubnetIds: Optional[StringList]
    SecurityGroupIds: Optional[StringList]


class SnapshotOptions(TypedDict, total=False):
    """Specifies the time, in UTC format, when the service takes a daily
    automated snapshot of the specified Elasticsearch domain. Default value
    is ``0`` hours.
    """

    AutomatedSnapshotStartHour: Optional[IntegerClass]


class EBSOptions(TypedDict, total=False):
    """Options to enable, disable, and specify the properties of EBS storage
    volumes. For more information, see `Configuring EBS-based
    Storage <http://docs.aws.amazon.com/elasticsearch-service/latest/developerguide/es-createupdatedomains.html#es-createdomain-configure-ebs>`__.
    """

    EBSEnabled: Optional[Boolean]
    VolumeType: Optional[VolumeType]
    VolumeSize: Optional[IntegerClass]
    Iops: Optional[IntegerClass]


class ZoneAwarenessConfig(TypedDict, total=False):
    """Specifies the zone awareness configuration for the domain cluster, such
    as the number of availability zones.
    """

    AvailabilityZoneCount: Optional[IntegerClass]


class ElasticsearchClusterConfig(TypedDict, total=False):
    """Specifies the configuration for the domain cluster, such as the type and
    number of instances.
    """

    InstanceType: Optional[ESPartitionInstanceType]
    InstanceCount: Optional[IntegerClass]
    DedicatedMasterEnabled: Optional[Boolean]
    ZoneAwarenessEnabled: Optional[Boolean]
    ZoneAwarenessConfig: Optional[ZoneAwarenessConfig]
    DedicatedMasterType: Optional[ESPartitionInstanceType]
    DedicatedMasterCount: Optional[IntegerClass]
    WarmEnabled: Optional[Boolean]
    WarmType: Optional[ESWarmPartitionInstanceType]
    WarmCount: Optional[IntegerClass]
    ColdStorageOptions: Optional[ColdStorageOptions]


class CreateElasticsearchDomainRequest(ServiceRequest):
    DomainName: DomainName
    ElasticsearchVersion: Optional[ElasticsearchVersionString]
    ElasticsearchClusterConfig: Optional[ElasticsearchClusterConfig]
    EBSOptions: Optional[EBSOptions]
    AccessPolicies: Optional[PolicyDocument]
    SnapshotOptions: Optional[SnapshotOptions]
    VPCOptions: Optional[VPCOptions]
    CognitoOptions: Optional[CognitoOptions]
    EncryptionAtRestOptions: Optional[EncryptionAtRestOptions]
    NodeToNodeEncryptionOptions: Optional[NodeToNodeEncryptionOptions]
    AdvancedOptions: Optional[AdvancedOptions]
    LogPublishingOptions: Optional[LogPublishingOptions]
    DomainEndpointOptions: Optional[DomainEndpointOptions]
    AdvancedSecurityOptions: Optional[AdvancedSecurityOptionsInput]
    AutoTuneOptions: Optional[AutoTuneOptionsInput]
    TagList: Optional[TagList]


class VPCDerivedInfo(TypedDict, total=False):
    """Options to specify the subnets and security groups for VPC endpoint. For
    more information, see `VPC Endpoints for Amazon Elasticsearch Service
    Domains <http://docs.aws.amazon.com/elasticsearch-service/latest/developerguide/es-vpc.html>`__.
    """

    VPCId: Optional[String]
    SubnetIds: Optional[StringList]
    AvailabilityZones: Optional[StringList]
    SecurityGroupIds: Optional[StringList]


EndpointsMap = Dict[String, ServiceUrl]


class ElasticsearchDomainStatus(TypedDict, total=False):
    """The current status of an Elasticsearch domain."""

    DomainId: DomainId
    DomainName: DomainName
    ARN: ARN
    Created: Optional[Boolean]
    Deleted: Optional[Boolean]
    Endpoint: Optional[ServiceUrl]
    Endpoints: Optional[EndpointsMap]
    Processing: Optional[Boolean]
    UpgradeProcessing: Optional[Boolean]
    ElasticsearchVersion: Optional[ElasticsearchVersionString]
    ElasticsearchClusterConfig: ElasticsearchClusterConfig
    EBSOptions: Optional[EBSOptions]
    AccessPolicies: Optional[PolicyDocument]
    SnapshotOptions: Optional[SnapshotOptions]
    VPCOptions: Optional[VPCDerivedInfo]
    CognitoOptions: Optional[CognitoOptions]
    EncryptionAtRestOptions: Optional[EncryptionAtRestOptions]
    NodeToNodeEncryptionOptions: Optional[NodeToNodeEncryptionOptions]
    AdvancedOptions: Optional[AdvancedOptions]
    LogPublishingOptions: Optional[LogPublishingOptions]
    ServiceSoftwareOptions: Optional[ServiceSoftwareOptions]
    DomainEndpointOptions: Optional[DomainEndpointOptions]
    AdvancedSecurityOptions: Optional[AdvancedSecurityOptions]
    AutoTuneOptions: Optional[AutoTuneOptionsOutput]


class CreateElasticsearchDomainResponse(TypedDict, total=False):
    """The result of a ``CreateElasticsearchDomain`` operation. Contains the
    status of the newly created Elasticsearch domain.
    """

    DomainStatus: Optional[ElasticsearchDomainStatus]


class CreateOutboundCrossClusterSearchConnectionRequest(ServiceRequest):
    """Container for the parameters to the
    ``CreateOutboundCrossClusterSearchConnection`` operation.
    """

    SourceDomainInfo: DomainInformation
    DestinationDomainInfo: DomainInformation
    ConnectionAlias: ConnectionAlias


class OutboundCrossClusterSearchConnectionStatus(TypedDict, total=False):
    """Specifies the connection status of an outbound cross-cluster search
    connection.
    """

    StatusCode: Optional[OutboundCrossClusterSearchConnectionStatusCode]
    Message: Optional[CrossClusterSearchConnectionStatusMessage]


class CreateOutboundCrossClusterSearchConnectionResponse(TypedDict, total=False):
    """The result of a ``CreateOutboundCrossClusterSearchConnection`` request.
    Contains the details of the newly created cross-cluster search
    connection.
    """

    SourceDomainInfo: Optional[DomainInformation]
    DestinationDomainInfo: Optional[DomainInformation]
    ConnectionAlias: Optional[ConnectionAlias]
    ConnectionStatus: Optional[OutboundCrossClusterSearchConnectionStatus]
    CrossClusterSearchConnectionId: Optional[CrossClusterSearchConnectionId]


class PackageSource(TypedDict, total=False):
    """The S3 location for importing the package specified as ``S3BucketName``
    and ``S3Key``
    """

    S3BucketName: Optional[S3BucketName]
    S3Key: Optional[S3Key]


class CreatePackageRequest(ServiceRequest):
    """Container for request parameters to ``CreatePackage`` operation."""

    PackageName: PackageName
    PackageType: PackageType
    PackageDescription: Optional[PackageDescription]
    PackageSource: PackageSource


CreatedAt = datetime


class PackageDetails(TypedDict, total=False):
    """Basic information about a package."""

    PackageID: Optional[PackageID]
    PackageName: Optional[PackageName]
    PackageType: Optional[PackageType]
    PackageDescription: Optional[PackageDescription]
    PackageStatus: Optional[PackageStatus]
    CreatedAt: Optional[CreatedAt]
    LastUpdatedAt: Optional[LastUpdated]
    AvailablePackageVersion: Optional[PackageVersion]
    ErrorDetails: Optional[ErrorDetails]


class CreatePackageResponse(TypedDict, total=False):
    """Container for response returned by ``CreatePackage`` operation."""

    PackageDetails: Optional[PackageDetails]


class DeleteElasticsearchDomainRequest(ServiceRequest):
    """Container for the parameters to the ``DeleteElasticsearchDomain``
    operation. Specifies the name of the Elasticsearch domain that you want
    to delete.
    """

    DomainName: DomainName


class DeleteElasticsearchDomainResponse(TypedDict, total=False):
    """The result of a ``DeleteElasticsearchDomain`` request. Contains the
    status of the pending deletion, or no status if the domain and all of
    its resources have been deleted.
    """

    DomainStatus: Optional[ElasticsearchDomainStatus]


class DeleteInboundCrossClusterSearchConnectionRequest(ServiceRequest):
    """Container for the parameters to the
    ``DeleteInboundCrossClusterSearchConnection`` operation.
    """

    CrossClusterSearchConnectionId: CrossClusterSearchConnectionId


class DeleteInboundCrossClusterSearchConnectionResponse(TypedDict, total=False):
    """The result of a ``DeleteInboundCrossClusterSearchConnection`` operation.
    Contains details of deleted inbound connection.
    """

    CrossClusterSearchConnection: Optional[InboundCrossClusterSearchConnection]


class DeleteOutboundCrossClusterSearchConnectionRequest(ServiceRequest):
    """Container for the parameters to the
    ``DeleteOutboundCrossClusterSearchConnection`` operation.
    """

    CrossClusterSearchConnectionId: CrossClusterSearchConnectionId


class OutboundCrossClusterSearchConnection(TypedDict, total=False):
    """Specifies details of an outbound connection."""

    SourceDomainInfo: Optional[DomainInformation]
    DestinationDomainInfo: Optional[DomainInformation]
    CrossClusterSearchConnectionId: Optional[CrossClusterSearchConnectionId]
    ConnectionAlias: Optional[ConnectionAlias]
    ConnectionStatus: Optional[OutboundCrossClusterSearchConnectionStatus]


class DeleteOutboundCrossClusterSearchConnectionResponse(TypedDict, total=False):
    """The result of a ``DeleteOutboundCrossClusterSearchConnection``
    operation. Contains details of deleted outbound connection.
    """

    CrossClusterSearchConnection: Optional[OutboundCrossClusterSearchConnection]


class DeletePackageRequest(ServiceRequest):
    """Container for request parameters to ``DeletePackage`` operation."""

    PackageID: PackageID


class DeletePackageResponse(TypedDict, total=False):
    """Container for response parameters to ``DeletePackage`` operation."""

    PackageDetails: Optional[PackageDetails]


class DescribeDomainAutoTunesRequest(ServiceRequest):
    """Container for the parameters to the ``DescribeDomainAutoTunes``
    operation.
    """

    DomainName: DomainName
    MaxResults: Optional[MaxResults]
    NextToken: Optional[NextToken]


class DescribeDomainAutoTunesResponse(TypedDict, total=False):
    """The result of ``DescribeDomainAutoTunes`` request. See the `Developer
    Guide <https://docs.aws.amazon.com/elasticsearch-service/latest/developerguide/auto-tune.html>`__
    for more information.
    """

    AutoTunes: Optional[AutoTuneList]
    NextToken: Optional[NextToken]


class DescribeElasticsearchDomainConfigRequest(ServiceRequest):
    """Container for the parameters to the
    ``DescribeElasticsearchDomainConfig`` operation. Specifies the domain
    name for which you want configuration information.
    """

    DomainName: DomainName


class DomainEndpointOptionsStatus(TypedDict, total=False):
    """The configured endpoint options for the domain and their current status."""

    Options: DomainEndpointOptions
    Status: OptionStatus


class LogPublishingOptionsStatus(TypedDict, total=False):
    """The configured log publishing options for the domain and their current
    status.
    """

    Options: Optional[LogPublishingOptions]
    Status: Optional[OptionStatus]


class NodeToNodeEncryptionOptionsStatus(TypedDict, total=False):
    """Status of the node-to-node encryption options for the specified
    Elasticsearch domain.
    """

    Options: NodeToNodeEncryptionOptions
    Status: OptionStatus


class EncryptionAtRestOptionsStatus(TypedDict, total=False):
    """Status of the Encryption At Rest options for the specified Elasticsearch
    domain.
    """

    Options: EncryptionAtRestOptions
    Status: OptionStatus


class VPCDerivedInfoStatus(TypedDict, total=False):
    """Status of the VPC options for the specified Elasticsearch domain."""

    Options: VPCDerivedInfo
    Status: OptionStatus


class SnapshotOptionsStatus(TypedDict, total=False):
    """Status of a daily automated snapshot."""

    Options: SnapshotOptions
    Status: OptionStatus


class EBSOptionsStatus(TypedDict, total=False):
    """Status of the EBS options for the specified Elasticsearch domain."""

    Options: EBSOptions
    Status: OptionStatus


class ElasticsearchClusterConfigStatus(TypedDict, total=False):
    """Specifies the configuration status for the specified Elasticsearch
    domain.
    """

    Options: ElasticsearchClusterConfig
    Status: OptionStatus


class ElasticsearchVersionStatus(TypedDict, total=False):
    """Status of the Elasticsearch version options for the specified
    Elasticsearch domain.
    """

    Options: ElasticsearchVersionString
    Status: OptionStatus


class ElasticsearchDomainConfig(TypedDict, total=False):
    """The configuration of an Elasticsearch domain."""

    ElasticsearchVersion: Optional[ElasticsearchVersionStatus]
    ElasticsearchClusterConfig: Optional[ElasticsearchClusterConfigStatus]
    EBSOptions: Optional[EBSOptionsStatus]
    AccessPolicies: Optional[AccessPoliciesStatus]
    SnapshotOptions: Optional[SnapshotOptionsStatus]
    VPCOptions: Optional[VPCDerivedInfoStatus]
    CognitoOptions: Optional[CognitoOptionsStatus]
    EncryptionAtRestOptions: Optional[EncryptionAtRestOptionsStatus]
    NodeToNodeEncryptionOptions: Optional[NodeToNodeEncryptionOptionsStatus]
    AdvancedOptions: Optional[AdvancedOptionsStatus]
    LogPublishingOptions: Optional[LogPublishingOptionsStatus]
    DomainEndpointOptions: Optional[DomainEndpointOptionsStatus]
    AdvancedSecurityOptions: Optional[AdvancedSecurityOptionsStatus]
    AutoTuneOptions: Optional[AutoTuneOptionsStatus]


class DescribeElasticsearchDomainConfigResponse(TypedDict, total=False):
    """The result of a ``DescribeElasticsearchDomainConfig`` request. Contains
    the configuration information of the requested domain.
    """

    DomainConfig: ElasticsearchDomainConfig


class DescribeElasticsearchDomainRequest(ServiceRequest):
    """Container for the parameters to the ``DescribeElasticsearchDomain``
    operation.
    """

    DomainName: DomainName


class DescribeElasticsearchDomainResponse(TypedDict, total=False):
    """The result of a ``DescribeElasticsearchDomain`` request. Contains the
    status of the domain specified in the request.
    """

    DomainStatus: ElasticsearchDomainStatus


DomainNameList = List[DomainName]


class DescribeElasticsearchDomainsRequest(ServiceRequest):
    """Container for the parameters to the ``DescribeElasticsearchDomains``
    operation. By default, the API returns the status of all Elasticsearch
    domains.
    """

    DomainNames: DomainNameList


ElasticsearchDomainStatusList = List[ElasticsearchDomainStatus]


class DescribeElasticsearchDomainsResponse(TypedDict, total=False):
    """The result of a ``DescribeElasticsearchDomains`` request. Contains the
    status of the specified domains or all domains owned by the account.
    """

    DomainStatusList: ElasticsearchDomainStatusList


class DescribeElasticsearchInstanceTypeLimitsRequest(ServiceRequest):
    """Container for the parameters to
    ``DescribeElasticsearchInstanceTypeLimits`` operation.
    """

    DomainName: Optional[DomainName]
    InstanceType: ESPartitionInstanceType
    ElasticsearchVersion: ElasticsearchVersionString


class InstanceCountLimits(TypedDict, total=False):
    """InstanceCountLimits represents the limits on number of instances that be
    created in Amazon Elasticsearch for given InstanceType.
    """

    MinimumInstanceCount: Optional[MinimumInstanceCount]
    MaximumInstanceCount: Optional[MaximumInstanceCount]


class InstanceLimits(TypedDict, total=False):
    """InstanceLimits represents the list of instance related attributes that
    are available for given InstanceType.
    """

    InstanceCountLimits: Optional[InstanceCountLimits]


class StorageTypeLimit(TypedDict, total=False):
    """Limits that are applicable for given storage type."""

    LimitName: Optional[LimitName]
    LimitValues: Optional[LimitValueList]


StorageTypeLimitList = List[StorageTypeLimit]


class StorageType(TypedDict, total=False):
    """StorageTypes represents the list of storage related types and their
    attributes that are available for given InstanceType.
    """

    StorageTypeName: Optional[StorageTypeName]
    StorageSubTypeName: Optional[StorageSubTypeName]
    StorageTypeLimits: Optional[StorageTypeLimitList]


StorageTypeList = List[StorageType]


class Limits(TypedDict, total=False):
    """| Limits for given InstanceType and for each of it's role.
    | Limits contains following ``StorageTypes,`` ``InstanceLimits`` and
      ``AdditionalLimits``
    """

    StorageTypes: Optional[StorageTypeList]
    InstanceLimits: Optional[InstanceLimits]
    AdditionalLimits: Optional[AdditionalLimitList]


LimitsByRole = Dict[InstanceRole, Limits]


class DescribeElasticsearchInstanceTypeLimitsResponse(TypedDict, total=False):
    """Container for the parameters received from
    ``DescribeElasticsearchInstanceTypeLimits`` operation.
    """

    LimitsByRole: Optional[LimitsByRole]


ValueStringList = List[NonEmptyString]


class Filter(TypedDict, total=False):
    """A filter used to limit results when describing inbound or outbound
    cross-cluster search connections. Multiple values can be specified per
    filter. A cross-cluster search connection must match at least one of the
    specified values for it to be returned from an operation.
    """

    Name: Optional[NonEmptyString]
    Values: Optional[ValueStringList]


FilterList = List[Filter]


class DescribeInboundCrossClusterSearchConnectionsRequest(ServiceRequest):
    """Container for the parameters to the
    ``DescribeInboundCrossClusterSearchConnections`` operation.
    """

    Filters: Optional[FilterList]
    MaxResults: Optional[MaxResults]
    NextToken: Optional[NextToken]


InboundCrossClusterSearchConnections = List[InboundCrossClusterSearchConnection]


class DescribeInboundCrossClusterSearchConnectionsResponse(TypedDict, total=False):
    """The result of a ``DescribeInboundCrossClusterSearchConnections``
    request. Contains the list of connections matching the filter criteria.
    """

    CrossClusterSearchConnections: Optional[InboundCrossClusterSearchConnections]
    NextToken: Optional[NextToken]


class DescribeOutboundCrossClusterSearchConnectionsRequest(ServiceRequest):
    """Container for the parameters to the
    ``DescribeOutboundCrossClusterSearchConnections`` operation.
    """

    Filters: Optional[FilterList]
    MaxResults: Optional[MaxResults]
    NextToken: Optional[NextToken]


OutboundCrossClusterSearchConnections = List[OutboundCrossClusterSearchConnection]


class DescribeOutboundCrossClusterSearchConnectionsResponse(TypedDict, total=False):
    """The result of a ``DescribeOutboundCrossClusterSearchConnections``
    request. Contains the list of connections matching the filter criteria.
    """

    CrossClusterSearchConnections: Optional[OutboundCrossClusterSearchConnections]
    NextToken: Optional[NextToken]


DescribePackagesFilterValues = List[DescribePackagesFilterValue]


class DescribePackagesFilter(TypedDict, total=False):
    """Filter to apply in ``DescribePackage`` response."""

    Name: Optional[DescribePackagesFilterName]
    Value: Optional[DescribePackagesFilterValues]


DescribePackagesFilterList = List[DescribePackagesFilter]


class DescribePackagesRequest(ServiceRequest):
    """Container for request parameters to ``DescribePackage`` operation."""

    Filters: Optional[DescribePackagesFilterList]
    MaxResults: Optional[MaxResults]
    NextToken: Optional[NextToken]


PackageDetailsList = List[PackageDetails]


class DescribePackagesResponse(TypedDict, total=False):
    """Container for response returned by ``DescribePackages`` operation."""

    PackageDetailsList: Optional[PackageDetailsList]
    NextToken: Optional[String]


class DescribeReservedElasticsearchInstanceOfferingsRequest(ServiceRequest):
    """Container for parameters to
    ``DescribeReservedElasticsearchInstanceOfferings``
    """

    ReservedElasticsearchInstanceOfferingId: Optional[GUID]
    MaxResults: Optional[MaxResults]
    NextToken: Optional[NextToken]


class RecurringCharge(TypedDict, total=False):
    """Contains the specific price and frequency of a recurring charges for a
    reserved Elasticsearch instance, or for a reserved Elasticsearch
    instance offering.
    """

    RecurringChargeAmount: Optional[Double]
    RecurringChargeFrequency: Optional[String]


RecurringChargeList = List[RecurringCharge]


class ReservedElasticsearchInstanceOffering(TypedDict, total=False):
    """Details of a reserved Elasticsearch instance offering."""

    ReservedElasticsearchInstanceOfferingId: Optional[GUID]
    ElasticsearchInstanceType: Optional[ESPartitionInstanceType]
    Duration: Optional[Integer]
    FixedPrice: Optional[Double]
    UsagePrice: Optional[Double]
    CurrencyCode: Optional[String]
    PaymentOption: Optional[ReservedElasticsearchInstancePaymentOption]
    RecurringCharges: Optional[RecurringChargeList]


ReservedElasticsearchInstanceOfferingList = List[ReservedElasticsearchInstanceOffering]


class DescribeReservedElasticsearchInstanceOfferingsResponse(TypedDict, total=False):
    """Container for results from
    ``DescribeReservedElasticsearchInstanceOfferings``
    """

    NextToken: Optional[NextToken]
    ReservedElasticsearchInstanceOfferings: Optional[ReservedElasticsearchInstanceOfferingList]


class DescribeReservedElasticsearchInstancesRequest(ServiceRequest):
    """Container for parameters to ``DescribeReservedElasticsearchInstances``"""

    ReservedElasticsearchInstanceId: Optional[GUID]
    MaxResults: Optional[MaxResults]
    NextToken: Optional[NextToken]


class ReservedElasticsearchInstance(TypedDict, total=False):
    """Details of a reserved Elasticsearch instance."""

    ReservationName: Optional[ReservationToken]
    ReservedElasticsearchInstanceId: Optional[GUID]
    ReservedElasticsearchInstanceOfferingId: Optional[String]
    ElasticsearchInstanceType: Optional[ESPartitionInstanceType]
    StartTime: Optional[UpdateTimestamp]
    Duration: Optional[Integer]
    FixedPrice: Optional[Double]
    UsagePrice: Optional[Double]
    CurrencyCode: Optional[String]
    ElasticsearchInstanceCount: Optional[Integer]
    State: Optional[String]
    PaymentOption: Optional[ReservedElasticsearchInstancePaymentOption]
    RecurringCharges: Optional[RecurringChargeList]


ReservedElasticsearchInstanceList = List[ReservedElasticsearchInstance]


class DescribeReservedElasticsearchInstancesResponse(TypedDict, total=False):
    """Container for results from ``DescribeReservedElasticsearchInstances``"""

    NextToken: Optional[String]
    ReservedElasticsearchInstances: Optional[ReservedElasticsearchInstanceList]


class DissociatePackageRequest(ServiceRequest):
    """Container for request parameters to ``DissociatePackage`` operation."""

    PackageID: PackageID
    DomainName: DomainName


class DissociatePackageResponse(TypedDict, total=False):
    """Container for response returned by ``DissociatePackage`` operation."""

    DomainPackageDetails: Optional[DomainPackageDetails]


class DomainInfo(TypedDict, total=False):
    DomainName: Optional[DomainName]
    EngineType: Optional[EngineType]


DomainInfoList = List[DomainInfo]
DomainPackageDetailsList = List[DomainPackageDetails]


class DryRunResults(TypedDict, total=False):
    DeploymentType: Optional[DeploymentType]
    Message: Optional[Message]


ElasticsearchInstanceTypeList = List[ESPartitionInstanceType]


class GetCompatibleElasticsearchVersionsRequest(ServiceRequest):
    """Container for request parameters to
    ``GetCompatibleElasticsearchVersions`` operation.
    """

    DomainName: Optional[DomainName]


class GetCompatibleElasticsearchVersionsResponse(TypedDict, total=False):
    """Container for response returned by
    ``GetCompatibleElasticsearchVersions`` operation.
    """

    CompatibleElasticsearchVersions: Optional[CompatibleElasticsearchVersionsList]


class GetPackageVersionHistoryRequest(ServiceRequest):
    """Container for request parameters to ``GetPackageVersionHistory``
    operation.
    """

    PackageID: PackageID
    MaxResults: Optional[MaxResults]
    NextToken: Optional[NextToken]


class PackageVersionHistory(TypedDict, total=False):
    """Details of a package version."""

    PackageVersion: Optional[PackageVersion]
    CommitMessage: Optional[CommitMessage]
    CreatedAt: Optional[CreatedAt]


PackageVersionHistoryList = List[PackageVersionHistory]


class GetPackageVersionHistoryResponse(TypedDict, total=False):
    """Container for response returned by ``GetPackageVersionHistory``
    operation.
    """

    PackageID: Optional[PackageID]
    PackageVersionHistoryList: Optional[PackageVersionHistoryList]
    NextToken: Optional[String]


class GetUpgradeHistoryRequest(ServiceRequest):
    """Container for request parameters to ``GetUpgradeHistory`` operation."""

    DomainName: DomainName
    MaxResults: Optional[MaxResults]
    NextToken: Optional[NextToken]


Issues = List[Issue]


class UpgradeStepItem(TypedDict, total=False):
    """Represents a single step of the Upgrade or Upgrade Eligibility Check
    workflow.
    """

    UpgradeStep: Optional[UpgradeStep]
    UpgradeStepStatus: Optional[UpgradeStatus]
    Issues: Optional[Issues]
    ProgressPercent: Optional[Double]


UpgradeStepsList = List[UpgradeStepItem]
StartTimestamp = datetime


class UpgradeHistory(TypedDict, total=False):
    """History of the last 10 Upgrades and Upgrade Eligibility Checks."""

    UpgradeName: Optional[UpgradeName]
    StartTimestamp: Optional[StartTimestamp]
    UpgradeStatus: Optional[UpgradeStatus]
    StepsList: Optional[UpgradeStepsList]


UpgradeHistoryList = List[UpgradeHistory]


class GetUpgradeHistoryResponse(TypedDict, total=False):
    """Container for response returned by ``GetUpgradeHistory`` operation."""

    UpgradeHistories: Optional[UpgradeHistoryList]
    NextToken: Optional[String]


class GetUpgradeStatusRequest(ServiceRequest):
    """Container for request parameters to ``GetUpgradeStatus`` operation."""

    DomainName: DomainName


class GetUpgradeStatusResponse(TypedDict, total=False):
    """Container for response returned by ``GetUpgradeStatus`` operation."""

    UpgradeStep: Optional[UpgradeStep]
    StepStatus: Optional[UpgradeStatus]
    UpgradeName: Optional[UpgradeName]


class ListDomainNamesRequest(ServiceRequest):
    """Container for the parameters to the ``ListDomainNames`` operation."""

    EngineType: Optional[EngineType]


class ListDomainNamesResponse(TypedDict, total=False):
    """The result of a ``ListDomainNames`` operation. Contains the names of all
    domains owned by this account and their respective engine types.
    """

    DomainNames: Optional[DomainInfoList]


class ListDomainsForPackageRequest(ServiceRequest):
    """Container for request parameters to ``ListDomainsForPackage`` operation."""

    PackageID: PackageID
    MaxResults: Optional[MaxResults]
    NextToken: Optional[NextToken]


class ListDomainsForPackageResponse(TypedDict, total=False):
    """Container for response parameters to ``ListDomainsForPackage``
    operation.
    """

    DomainPackageDetailsList: Optional[DomainPackageDetailsList]
    NextToken: Optional[String]


class ListElasticsearchInstanceTypesRequest(ServiceRequest):
    """Container for the parameters to the ``ListElasticsearchInstanceTypes``
    operation.
    """

    ElasticsearchVersion: ElasticsearchVersionString
    DomainName: Optional[DomainName]
    MaxResults: Optional[MaxResults]
    NextToken: Optional[NextToken]


class ListElasticsearchInstanceTypesResponse(TypedDict, total=False):
    """Container for the parameters returned by
    ``ListElasticsearchInstanceTypes`` operation.
    """

    ElasticsearchInstanceTypes: Optional[ElasticsearchInstanceTypeList]
    NextToken: Optional[NextToken]


class ListElasticsearchVersionsRequest(ServiceRequest):
    """Container for the parameters to the ``ListElasticsearchVersions``
    operation.

    Use ``MaxResults`` to control the maximum number of results to retrieve
    in a single call.

    Use ``NextToken`` in response to retrieve more results. If the received
    response does not contain a NextToken, then there are no more results to
    retrieve.
    """

    MaxResults: Optional[MaxResults]
    NextToken: Optional[NextToken]


class ListElasticsearchVersionsResponse(TypedDict, total=False):
    """Container for the parameters for response received from
    ``ListElasticsearchVersions`` operation.
    """

    ElasticsearchVersions: Optional[ElasticsearchVersionList]
    NextToken: Optional[NextToken]


class ListPackagesForDomainRequest(ServiceRequest):
    """Container for request parameters to ``ListPackagesForDomain`` operation."""

    DomainName: DomainName
    MaxResults: Optional[MaxResults]
    NextToken: Optional[NextToken]


class ListPackagesForDomainResponse(TypedDict, total=False):
    """Container for response parameters to ``ListPackagesForDomain``
    operation.
    """

    DomainPackageDetailsList: Optional[DomainPackageDetailsList]
    NextToken: Optional[String]


class ListTagsRequest(ServiceRequest):
    """Container for the parameters to the ``ListTags`` operation. Specify the
    ``ARN`` for the Elasticsearch domain to which the tags are attached that
    you want to view are attached.
    """

    ARN: ARN


class ListTagsResponse(TypedDict, total=False):
    """The result of a ``ListTags`` operation. Contains tags for all requested
    Elasticsearch domains.
    """

    TagList: Optional[TagList]


class PurchaseReservedElasticsearchInstanceOfferingRequest(ServiceRequest):
    """Container for parameters to
    ``PurchaseReservedElasticsearchInstanceOffering``
    """

    ReservedElasticsearchInstanceOfferingId: GUID
    ReservationName: ReservationToken
    InstanceCount: Optional[InstanceCount]


class PurchaseReservedElasticsearchInstanceOfferingResponse(TypedDict, total=False):
    """Represents the output of a
    ``PurchaseReservedElasticsearchInstanceOffering`` operation.
    """

    ReservedElasticsearchInstanceId: Optional[GUID]
    ReservationName: Optional[ReservationToken]


class RejectInboundCrossClusterSearchConnectionRequest(ServiceRequest):
    """Container for the parameters to the
    ``RejectInboundCrossClusterSearchConnection`` operation.
    """

    CrossClusterSearchConnectionId: CrossClusterSearchConnectionId


class RejectInboundCrossClusterSearchConnectionResponse(TypedDict, total=False):
    """The result of a ``RejectInboundCrossClusterSearchConnection`` operation.
    Contains details of rejected inbound connection.
    """

    CrossClusterSearchConnection: Optional[InboundCrossClusterSearchConnection]


class RemoveTagsRequest(ServiceRequest):
    """Container for the parameters to the ``RemoveTags`` operation. Specify
    the ``ARN`` for the Elasticsearch domain from which you want to remove
    the specified ``TagKey``.
    """

    ARN: ARN
    TagKeys: StringList


class StartElasticsearchServiceSoftwareUpdateRequest(ServiceRequest):
    """Container for the parameters to the
    ``StartElasticsearchServiceSoftwareUpdate`` operation. Specifies the
    name of the Elasticsearch domain that you wish to schedule a service
    software update on.
    """

    DomainName: DomainName


class StartElasticsearchServiceSoftwareUpdateResponse(TypedDict, total=False):
    """The result of a ``StartElasticsearchServiceSoftwareUpdate`` operation.
    Contains the status of the update.
    """

    ServiceSoftwareOptions: Optional[ServiceSoftwareOptions]


class UpdateElasticsearchDomainConfigRequest(ServiceRequest):
    """Container for the parameters to the ``UpdateElasticsearchDomain``
    operation. Specifies the type and number of instances in the domain
    cluster.
    """

    DomainName: DomainName
    ElasticsearchClusterConfig: Optional[ElasticsearchClusterConfig]
    EBSOptions: Optional[EBSOptions]
    SnapshotOptions: Optional[SnapshotOptions]
    VPCOptions: Optional[VPCOptions]
    CognitoOptions: Optional[CognitoOptions]
    AdvancedOptions: Optional[AdvancedOptions]
    AccessPolicies: Optional[PolicyDocument]
    LogPublishingOptions: Optional[LogPublishingOptions]
    DomainEndpointOptions: Optional[DomainEndpointOptions]
    AdvancedSecurityOptions: Optional[AdvancedSecurityOptionsInput]
    NodeToNodeEncryptionOptions: Optional[NodeToNodeEncryptionOptions]
    EncryptionAtRestOptions: Optional[EncryptionAtRestOptions]
    AutoTuneOptions: Optional[AutoTuneOptions]
    DryRun: Optional[DryRun]


class UpdateElasticsearchDomainConfigResponse(TypedDict, total=False):
    """The result of an ``UpdateElasticsearchDomain`` request. Contains the
    status of the Elasticsearch domain being updated.
    """

    DomainConfig: ElasticsearchDomainConfig
    DryRunResults: Optional[DryRunResults]


class UpdatePackageRequest(ServiceRequest):
    """Container for request parameters to ``UpdatePackage`` operation."""

    PackageID: PackageID
    PackageSource: PackageSource
    PackageDescription: Optional[PackageDescription]
    CommitMessage: Optional[CommitMessage]


class UpdatePackageResponse(TypedDict, total=False):
    """Container for response returned by ``UpdatePackage`` operation."""

    PackageDetails: Optional[PackageDetails]


class UpgradeElasticsearchDomainRequest(ServiceRequest):
    """Container for request parameters to ``UpgradeElasticsearchDomain``
    operation.
    """

    DomainName: DomainName
    TargetVersion: ElasticsearchVersionString
    PerformCheckOnly: Optional[Boolean]


class UpgradeElasticsearchDomainResponse(TypedDict, total=False):
    """Container for response returned by ``UpgradeElasticsearchDomain``
    operation.
    """

    DomainName: Optional[DomainName]
    TargetVersion: Optional[ElasticsearchVersionString]
    PerformCheckOnly: Optional[Boolean]


class EsApi:

    service = "es"
    version = "2015-01-01"

    @handler("AcceptInboundCrossClusterSearchConnection")
    def accept_inbound_cross_cluster_search_connection(
        self,
        context: RequestContext,
        cross_cluster_search_connection_id: CrossClusterSearchConnectionId,
    ) -> AcceptInboundCrossClusterSearchConnectionResponse:
        """Allows the destination domain owner to accept an inbound cross-cluster
        search connection request.

        :param cross_cluster_search_connection_id: The id of the inbound connection that you want to accept.
        :returns: AcceptInboundCrossClusterSearchConnectionResponse
        :raises ResourceNotFoundException:
        :raises LimitExceededException:
        :raises DisabledOperationException:
        """
        raise NotImplementedError

    @handler("AddTags")
    def add_tags(self, context: RequestContext, arn: ARN, tag_list: TagList) -> None:
        """Attaches tags to an existing Elasticsearch domain. Tags are a set of
        case-sensitive key value pairs. An Elasticsearch domain may have up to
        10 tags. See `Tagging Amazon Elasticsearch Service Domains for more
        information. <http://docs.aws.amazon.com/elasticsearch-service/latest/developerguide/es-managedomains.html#es-managedomains-awsresorcetagging>`__

        :param arn: Specify the ``ARN`` for which you want to add the tags.
        :param tag_list: List of ``Tag`` that need to be added for the Elasticsearch domain.
        :raises BaseException:
        :raises LimitExceededException:
        :raises ValidationException:
        :raises InternalException:
        """
        raise NotImplementedError

    @handler("AssociatePackage")
    def associate_package(
        self, context: RequestContext, package_id: PackageID, domain_name: DomainName
    ) -> AssociatePackageResponse:
        """Associates a package with an Amazon ES domain.

        :param package_id: Internal ID of the package that you want to associate with a domain.
        :param domain_name: Name of the domain that you want to associate the package with.
        :returns: AssociatePackageResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises AccessDeniedException:
        :raises ValidationException:
        :raises ConflictException:
        """
        raise NotImplementedError

    @handler("CancelElasticsearchServiceSoftwareUpdate")
    def cancel_elasticsearch_service_software_update(
        self, context: RequestContext, domain_name: DomainName
    ) -> CancelElasticsearchServiceSoftwareUpdateResponse:
        """Cancels a scheduled service software update for an Amazon ES domain. You
        can only perform this operation before the ``AutomatedUpdateDate`` and
        when the ``UpdateStatus`` is in the ``PENDING_UPDATE`` state.

        :param domain_name: The name of the domain that you want to stop the latest service software
        update on.
        :returns: CancelElasticsearchServiceSoftwareUpdateResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("CreateElasticsearchDomain")
    def create_elasticsearch_domain(
        self,
        context: RequestContext,
        domain_name: DomainName,
        elasticsearch_version: ElasticsearchVersionString = None,
        elasticsearch_cluster_config: ElasticsearchClusterConfig = None,
        ebs_options: EBSOptions = None,
        access_policies: PolicyDocument = None,
        snapshot_options: SnapshotOptions = None,
        vpc_options: VPCOptions = None,
        cognito_options: CognitoOptions = None,
        encryption_at_rest_options: EncryptionAtRestOptions = None,
        node_to_node_encryption_options: NodeToNodeEncryptionOptions = None,
        advanced_options: AdvancedOptions = None,
        log_publishing_options: LogPublishingOptions = None,
        domain_endpoint_options: DomainEndpointOptions = None,
        advanced_security_options: AdvancedSecurityOptionsInput = None,
        auto_tune_options: AutoTuneOptionsInput = None,
        tag_list: TagList = None,
    ) -> CreateElasticsearchDomainResponse:
        """Creates a new Elasticsearch domain. For more information, see `Creating
        Elasticsearch
        Domains <http://docs.aws.amazon.com/elasticsearch-service/latest/developerguide/es-createupdatedomains.html#es-createdomains>`__
        in the *Amazon Elasticsearch Service Developer Guide*.

        :param domain_name: The name of the Elasticsearch domain that you are creating.
        :param elasticsearch_version: String of format X.
        :param elasticsearch_cluster_config: Configuration options for an Elasticsearch domain.
        :param ebs_options: Options to enable, disable and specify the type and size of EBS storage
        volumes.
        :param access_policies: IAM access policy as a JSON-formatted string.
        :param snapshot_options: Option to set time, in UTC format, of the daily automated snapshot.
        :param vpc_options: Options to specify the subnets and security groups for VPC endpoint.
        :param cognito_options: Options to specify the Cognito user and identity pools for Kibana
        authentication.
        :param encryption_at_rest_options: Specifies the Encryption At Rest Options.
        :param node_to_node_encryption_options: Specifies the NodeToNodeEncryptionOptions.
        :param advanced_options: Option to allow references to indices in an HTTP request body.
        :param log_publishing_options: Map of ``LogType`` and ``LogPublishingOption``, each containing options
        to publish a given type of Elasticsearch log.
        :param domain_endpoint_options: Options to specify configuration that will be applied to the domain
        endpoint.
        :param advanced_security_options: Specifies advanced security options.
        :param auto_tune_options: Specifies Auto-Tune options.
        :param tag_list: A list of ``Tag`` added during domain creation.
        :returns: CreateElasticsearchDomainResponse
        :raises BaseException:
        :raises DisabledOperationException:
        :raises InternalException:
        :raises InvalidTypeException:
        :raises LimitExceededException:
        :raises ResourceAlreadyExistsException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("CreateOutboundCrossClusterSearchConnection")
    def create_outbound_cross_cluster_search_connection(
        self,
        context: RequestContext,
        source_domain_info: DomainInformation,
        destination_domain_info: DomainInformation,
        connection_alias: ConnectionAlias,
    ) -> CreateOutboundCrossClusterSearchConnectionResponse:
        """Creates a new cross-cluster search connection from a source domain to a
        destination domain.

        :param source_domain_info: Specifies the ``DomainInformation`` for the source Elasticsearch domain.
        :param destination_domain_info: Specifies the ``DomainInformation`` for the destination Elasticsearch
        domain.
        :param connection_alias: Specifies the connection alias that will be used by the customer for
        this connection.
        :returns: CreateOutboundCrossClusterSearchConnectionResponse
        :raises LimitExceededException:
        :raises InternalException:
        :raises ResourceAlreadyExistsException:
        :raises DisabledOperationException:
        """
        raise NotImplementedError

    @handler("CreatePackage")
    def create_package(
        self,
        context: RequestContext,
        package_name: PackageName,
        package_type: PackageType,
        package_source: PackageSource,
        package_description: PackageDescription = None,
    ) -> CreatePackageResponse:
        """Create a package for use with Amazon ES domains.

        :param package_name: Unique identifier for the package.
        :param package_type: Type of package.
        :param package_source: The customer S3 location ``PackageSource`` for importing the package.
        :param package_description: Description of the package.
        :returns: CreatePackageResponse
        :raises BaseException:
        :raises InternalException:
        :raises LimitExceededException:
        :raises InvalidTypeException:
        :raises ResourceAlreadyExistsException:
        :raises AccessDeniedException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("DeleteElasticsearchDomain")
    def delete_elasticsearch_domain(
        self, context: RequestContext, domain_name: DomainName
    ) -> DeleteElasticsearchDomainResponse:
        """Permanently deletes the specified Elasticsearch domain and all of its
        data. Once a domain is deleted, it cannot be recovered.

        :param domain_name: The name of the Elasticsearch domain that you want to permanently
        delete.
        :returns: DeleteElasticsearchDomainResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("DeleteElasticsearchServiceRole")
    def delete_elasticsearch_service_role(
        self,
        context: RequestContext,
    ) -> None:
        """Deletes the service-linked role that Elasticsearch Service uses to
        manage and maintain VPC domains. Role deletion will fail if any existing
        VPC domains use the role. You must delete any such Elasticsearch domains
        before deleting the role. See `Deleting Elasticsearch Service
        Role <http://docs.aws.amazon.com/elasticsearch-service/latest/developerguide/es-vpc.html#es-enabling-slr>`__
        in *VPC Endpoints for Amazon Elasticsearch Service Domains*.

        :raises BaseException:
        :raises InternalException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("DeleteInboundCrossClusterSearchConnection")
    def delete_inbound_cross_cluster_search_connection(
        self,
        context: RequestContext,
        cross_cluster_search_connection_id: CrossClusterSearchConnectionId,
    ) -> DeleteInboundCrossClusterSearchConnectionResponse:
        """Allows the destination domain owner to delete an existing inbound
        cross-cluster search connection.

        :param cross_cluster_search_connection_id: The id of the inbound connection that you want to permanently delete.
        :returns: DeleteInboundCrossClusterSearchConnectionResponse
        :raises ResourceNotFoundException:
        :raises DisabledOperationException:
        """
        raise NotImplementedError

    @handler("DeleteOutboundCrossClusterSearchConnection")
    def delete_outbound_cross_cluster_search_connection(
        self,
        context: RequestContext,
        cross_cluster_search_connection_id: CrossClusterSearchConnectionId,
    ) -> DeleteOutboundCrossClusterSearchConnectionResponse:
        """Allows the source domain owner to delete an existing outbound
        cross-cluster search connection.

        :param cross_cluster_search_connection_id: The id of the outbound connection that you want to permanently delete.
        :returns: DeleteOutboundCrossClusterSearchConnectionResponse
        :raises ResourceNotFoundException:
        :raises DisabledOperationException:
        """
        raise NotImplementedError

    @handler("DeletePackage")
    def delete_package(
        self, context: RequestContext, package_id: PackageID
    ) -> DeletePackageResponse:
        """Delete the package.

        :param package_id: Internal ID of the package that you want to delete.
        :returns: DeletePackageResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises AccessDeniedException:
        :raises ValidationException:
        :raises ConflictException:
        """
        raise NotImplementedError

    @handler("DescribeDomainAutoTunes")
    def describe_domain_auto_tunes(
        self,
        context: RequestContext,
        domain_name: DomainName,
        max_results: MaxResults = None,
        next_token: NextToken = None,
    ) -> DescribeDomainAutoTunesResponse:
        """Provides scheduled Auto-Tune action details for the Elasticsearch
        domain, such as Auto-Tune action type, description, severity, and
        scheduled date.

        :param domain_name: Specifies the domain name for which you want Auto-Tune action details.
        :param max_results: Set this value to limit the number of results returned.
        :param next_token: NextToken is sent in case the earlier API call results contain the
        NextToken.
        :returns: DescribeDomainAutoTunesResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("DescribeElasticsearchDomain")
    def describe_elasticsearch_domain(
        self, context: RequestContext, domain_name: DomainName
    ) -> DescribeElasticsearchDomainResponse:
        """Returns domain configuration information about the specified
        Elasticsearch domain, including the domain ID, domain endpoint, and
        domain ARN.

        :param domain_name: The name of the Elasticsearch domain for which you want information.
        :returns: DescribeElasticsearchDomainResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("DescribeElasticsearchDomainConfig")
    def describe_elasticsearch_domain_config(
        self, context: RequestContext, domain_name: DomainName
    ) -> DescribeElasticsearchDomainConfigResponse:
        """Provides cluster configuration information about the specified
        Elasticsearch domain, such as the state, creation date, update version,
        and update date for cluster options.

        :param domain_name: The Elasticsearch domain that you want to get information about.
        :returns: DescribeElasticsearchDomainConfigResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("DescribeElasticsearchDomains")
    def describe_elasticsearch_domains(
        self, context: RequestContext, domain_names: DomainNameList
    ) -> DescribeElasticsearchDomainsResponse:
        """Returns domain configuration information about the specified
        Elasticsearch domains, including the domain ID, domain endpoint, and
        domain ARN.

        :param domain_names: The Elasticsearch domains for which you want information.
        :returns: DescribeElasticsearchDomainsResponse
        :raises BaseException:
        :raises InternalException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("DescribeElasticsearchInstanceTypeLimits")
    def describe_elasticsearch_instance_type_limits(
        self,
        context: RequestContext,
        instance_type: ESPartitionInstanceType,
        elasticsearch_version: ElasticsearchVersionString,
        domain_name: DomainName = None,
    ) -> DescribeElasticsearchInstanceTypeLimitsResponse:
        """Describe Elasticsearch Limits for a given InstanceType and
        ElasticsearchVersion. When modifying existing Domain, specify the
        ``DomainName`` to know what Limits are supported for modifying.

        :param instance_type: The instance type for an Elasticsearch cluster for which Elasticsearch
        ``Limits`` are needed.
        :param elasticsearch_version: Version of Elasticsearch for which ``Limits`` are needed.
        :param domain_name: DomainName represents the name of the Domain that we are trying to
        modify.
        :returns: DescribeElasticsearchInstanceTypeLimitsResponse
        :raises BaseException:
        :raises InternalException:
        :raises InvalidTypeException:
        :raises LimitExceededException:
        :raises ResourceNotFoundException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("DescribeInboundCrossClusterSearchConnections")
    def describe_inbound_cross_cluster_search_connections(
        self,
        context: RequestContext,
        filters: FilterList = None,
        max_results: MaxResults = None,
        next_token: NextToken = None,
    ) -> DescribeInboundCrossClusterSearchConnectionsResponse:
        """Lists all the inbound cross-cluster search connections for a destination
        domain.

        :param filters: A list of filters used to match properties for inbound cross-cluster
        search connection.
        :param max_results: Set this value to limit the number of results returned.
        :param next_token: NextToken is sent in case the earlier API call results contain the
        NextToken.
        :returns: DescribeInboundCrossClusterSearchConnectionsResponse
        :raises InvalidPaginationTokenException:
        :raises DisabledOperationException:
        """
        raise NotImplementedError

    @handler("DescribeOutboundCrossClusterSearchConnections")
    def describe_outbound_cross_cluster_search_connections(
        self,
        context: RequestContext,
        filters: FilterList = None,
        max_results: MaxResults = None,
        next_token: NextToken = None,
    ) -> DescribeOutboundCrossClusterSearchConnectionsResponse:
        """Lists all the outbound cross-cluster search connections for a source
        domain.

        :param filters: A list of filters used to match properties for outbound cross-cluster
        search connection.
        :param max_results: Set this value to limit the number of results returned.
        :param next_token: NextToken is sent in case the earlier API call results contain the
        NextToken.
        :returns: DescribeOutboundCrossClusterSearchConnectionsResponse
        :raises InvalidPaginationTokenException:
        :raises DisabledOperationException:
        """
        raise NotImplementedError

    @handler("DescribePackages")
    def describe_packages(
        self,
        context: RequestContext,
        filters: DescribePackagesFilterList = None,
        max_results: MaxResults = None,
        next_token: NextToken = None,
    ) -> DescribePackagesResponse:
        """Describes all packages available to Amazon ES. Includes options for
        filtering, limiting the number of results, and pagination.

        :param filters: Only returns packages that match the ``DescribePackagesFilterList``
        values.
        :param max_results: Limits results to a maximum number of packages.
        :param next_token: Used for pagination.
        :returns: DescribePackagesResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises AccessDeniedException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("DescribeReservedElasticsearchInstanceOfferings")
    def describe_reserved_elasticsearch_instance_offerings(
        self,
        context: RequestContext,
        reserved_elasticsearch_instance_offering_id: GUID = None,
        max_results: MaxResults = None,
        next_token: NextToken = None,
    ) -> DescribeReservedElasticsearchInstanceOfferingsResponse:
        """Lists available reserved Elasticsearch instance offerings.

        :param reserved_elasticsearch_instance_offering_id: The offering identifier filter value.
        :param max_results: Set this value to limit the number of results returned.
        :param next_token: NextToken should be sent in case if earlier API call produced result
        containing NextToken.
        :returns: DescribeReservedElasticsearchInstanceOfferingsResponse
        :raises ResourceNotFoundException:
        :raises ValidationException:
        :raises DisabledOperationException:
        :raises InternalException:
        """
        raise NotImplementedError

    @handler("DescribeReservedElasticsearchInstances")
    def describe_reserved_elasticsearch_instances(
        self,
        context: RequestContext,
        reserved_elasticsearch_instance_id: GUID = None,
        max_results: MaxResults = None,
        next_token: NextToken = None,
    ) -> DescribeReservedElasticsearchInstancesResponse:
        """Returns information about reserved Elasticsearch instances for this
        account.

        :param reserved_elasticsearch_instance_id: The reserved instance identifier filter value.
        :param max_results: Set this value to limit the number of results returned.
        :param next_token: NextToken should be sent in case if earlier API call produced result
        containing NextToken.
        :returns: DescribeReservedElasticsearchInstancesResponse
        :raises ResourceNotFoundException:
        :raises InternalException:
        :raises ValidationException:
        :raises DisabledOperationException:
        """
        raise NotImplementedError

    @handler("DissociatePackage")
    def dissociate_package(
        self, context: RequestContext, package_id: PackageID, domain_name: DomainName
    ) -> DissociatePackageResponse:
        """Dissociates a package from the Amazon ES domain.

        :param package_id: Internal ID of the package that you want to associate with a domain.
        :param domain_name: Name of the domain that you want to associate the package with.
        :returns: DissociatePackageResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises AccessDeniedException:
        :raises ValidationException:
        :raises ConflictException:
        """
        raise NotImplementedError

    @handler("GetCompatibleElasticsearchVersions")
    def get_compatible_elasticsearch_versions(
        self, context: RequestContext, domain_name: DomainName = None
    ) -> GetCompatibleElasticsearchVersionsResponse:
        """Returns a list of upgrade compatible Elastisearch versions. You can
        optionally pass a ``DomainName`` to get all upgrade compatible
        Elasticsearch versions for that specific domain.

        :param domain_name: The name of an Elasticsearch domain.
        :returns: GetCompatibleElasticsearchVersionsResponse
        :raises BaseException:
        :raises ResourceNotFoundException:
        :raises DisabledOperationException:
        :raises ValidationException:
        :raises InternalException:
        """
        raise NotImplementedError

    @handler("GetPackageVersionHistory")
    def get_package_version_history(
        self,
        context: RequestContext,
        package_id: PackageID,
        max_results: MaxResults = None,
        next_token: NextToken = None,
    ) -> GetPackageVersionHistoryResponse:
        """Returns a list of versions of the package, along with their creation
        time and commit message.

        :param package_id: Returns an audit history of versions of the package.
        :param max_results: Limits results to a maximum number of versions.
        :param next_token: Used for pagination.
        :returns: GetPackageVersionHistoryResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises AccessDeniedException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("GetUpgradeHistory")
    def get_upgrade_history(
        self,
        context: RequestContext,
        domain_name: DomainName,
        max_results: MaxResults = None,
        next_token: NextToken = None,
    ) -> GetUpgradeHistoryResponse:
        """Retrieves the complete history of the last 10 upgrades that were
        performed on the domain.

        :param domain_name: The name of an Elasticsearch domain.
        :param max_results: Set this value to limit the number of results returned.
        :param next_token: Paginated APIs accepts NextToken input to returns next page results and
        provides a NextToken output in the response which can be used by the
        client to retrieve more results.
        :returns: GetUpgradeHistoryResponse
        :raises BaseException:
        :raises ResourceNotFoundException:
        :raises DisabledOperationException:
        :raises ValidationException:
        :raises InternalException:
        """
        raise NotImplementedError

    @handler("GetUpgradeStatus")
    def get_upgrade_status(
        self, context: RequestContext, domain_name: DomainName
    ) -> GetUpgradeStatusResponse:
        """Retrieves the latest status of the last upgrade or upgrade eligibility
        check that was performed on the domain.

        :param domain_name: The name of an Elasticsearch domain.
        :returns: GetUpgradeStatusResponse
        :raises BaseException:
        :raises ResourceNotFoundException:
        :raises DisabledOperationException:
        :raises ValidationException:
        :raises InternalException:
        """
        raise NotImplementedError

    @handler("ListDomainNames")
    def list_domain_names(
        self, context: RequestContext, engine_type: EngineType = None
    ) -> ListDomainNamesResponse:
        """Returns the name of all Elasticsearch domains owned by the current
        user's account.

        :param engine_type: Optional parameter to filter the output by domain engine type.
        :returns: ListDomainNamesResponse
        :raises BaseException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("ListDomainsForPackage")
    def list_domains_for_package(
        self,
        context: RequestContext,
        package_id: PackageID,
        max_results: MaxResults = None,
        next_token: NextToken = None,
    ) -> ListDomainsForPackageResponse:
        """Lists all Amazon ES domains associated with the package.

        :param package_id: The package for which to list domains.
        :param max_results: Limits results to a maximum number of domains.
        :param next_token: Used for pagination.
        :returns: ListDomainsForPackageResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises AccessDeniedException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("ListElasticsearchInstanceTypes")
    def list_elasticsearch_instance_types(
        self,
        context: RequestContext,
        elasticsearch_version: ElasticsearchVersionString,
        domain_name: DomainName = None,
        max_results: MaxResults = None,
        next_token: NextToken = None,
    ) -> ListElasticsearchInstanceTypesResponse:
        """List all Elasticsearch instance types that are supported for given
        ElasticsearchVersion

        :param elasticsearch_version: Version of Elasticsearch for which list of supported elasticsearch
        instance types are needed.
        :param domain_name: DomainName represents the name of the Domain that we are trying to
        modify.
        :param max_results: Set this value to limit the number of results returned.
        :param next_token: NextToken should be sent in case if earlier API call produced result
        containing NextToken.
        :returns: ListElasticsearchInstanceTypesResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("ListElasticsearchVersions")
    def list_elasticsearch_versions(
        self,
        context: RequestContext,
        max_results: MaxResults = None,
        next_token: NextToken = None,
    ) -> ListElasticsearchVersionsResponse:
        """List all supported Elasticsearch versions

        :param max_results: Set this value to limit the number of results returned.
        :param next_token: Paginated APIs accepts NextToken input to returns next page results and
        provides a NextToken output in the response which can be used by the
        client to retrieve more results.
        :returns: ListElasticsearchVersionsResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("ListPackagesForDomain")
    def list_packages_for_domain(
        self,
        context: RequestContext,
        domain_name: DomainName,
        max_results: MaxResults = None,
        next_token: NextToken = None,
    ) -> ListPackagesForDomainResponse:
        """Lists all packages associated with the Amazon ES domain.

        :param domain_name: The name of the domain for which you want to list associated packages.
        :param max_results: Limits results to a maximum number of packages.
        :param next_token: Used for pagination.
        :returns: ListPackagesForDomainResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises AccessDeniedException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("ListTags")
    def list_tags(self, context: RequestContext, arn: ARN) -> ListTagsResponse:
        """Returns all tags for the given Elasticsearch domain.

        :param arn: Specify the ``ARN`` for the Elasticsearch domain to which the tags are
        attached that you want to view.
        :returns: ListTagsResponse
        :raises BaseException:
        :raises ResourceNotFoundException:
        :raises ValidationException:
        :raises InternalException:
        """
        raise NotImplementedError

    @handler("PurchaseReservedElasticsearchInstanceOffering")
    def purchase_reserved_elasticsearch_instance_offering(
        self,
        context: RequestContext,
        reserved_elasticsearch_instance_offering_id: GUID,
        reservation_name: ReservationToken,
        instance_count: InstanceCount = None,
    ) -> PurchaseReservedElasticsearchInstanceOfferingResponse:
        """Allows you to purchase reserved Elasticsearch instances.

        :param reserved_elasticsearch_instance_offering_id: The ID of the reserved Elasticsearch instance offering to purchase.
        :param reservation_name: A customer-specified identifier to track this reservation.
        :param instance_count: The number of Elasticsearch instances to reserve.
        :returns: PurchaseReservedElasticsearchInstanceOfferingResponse
        :raises ResourceNotFoundException:
        :raises ResourceAlreadyExistsException:
        :raises LimitExceededException:
        :raises DisabledOperationException:
        :raises ValidationException:
        :raises InternalException:
        """
        raise NotImplementedError

    @handler("RejectInboundCrossClusterSearchConnection")
    def reject_inbound_cross_cluster_search_connection(
        self,
        context: RequestContext,
        cross_cluster_search_connection_id: CrossClusterSearchConnectionId,
    ) -> RejectInboundCrossClusterSearchConnectionResponse:
        """Allows the destination domain owner to reject an inbound cross-cluster
        search connection request.

        :param cross_cluster_search_connection_id: The id of the inbound connection that you want to reject.
        :returns: RejectInboundCrossClusterSearchConnectionResponse
        :raises ResourceNotFoundException:
        :raises DisabledOperationException:
        """
        raise NotImplementedError

    @handler("RemoveTags")
    def remove_tags(self, context: RequestContext, arn: ARN, tag_keys: StringList) -> None:
        """Removes the specified set of tags from the specified Elasticsearch
        domain.

        :param arn: Specifies the ``ARN`` for the Elasticsearch domain from which you want
        to delete the specified tags.
        :param tag_keys: Specifies the ``TagKey`` list which you want to remove from the
        Elasticsearch domain.
        :raises BaseException:
        :raises ValidationException:
        :raises InternalException:
        """
        raise NotImplementedError

    @handler("StartElasticsearchServiceSoftwareUpdate")
    def start_elasticsearch_service_software_update(
        self, context: RequestContext, domain_name: DomainName
    ) -> StartElasticsearchServiceSoftwareUpdateResponse:
        """Schedules a service software update for an Amazon ES domain.

        :param domain_name: The name of the domain that you want to update to the latest service
        software.
        :returns: StartElasticsearchServiceSoftwareUpdateResponse
        :raises BaseException:
        :raises InternalException:
        :raises ResourceNotFoundException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("UpdateElasticsearchDomainConfig")
    def update_elasticsearch_domain_config(
        self,
        context: RequestContext,
        domain_name: DomainName,
        elasticsearch_cluster_config: ElasticsearchClusterConfig = None,
        ebs_options: EBSOptions = None,
        snapshot_options: SnapshotOptions = None,
        vpc_options: VPCOptions = None,
        cognito_options: CognitoOptions = None,
        advanced_options: AdvancedOptions = None,
        access_policies: PolicyDocument = None,
        log_publishing_options: LogPublishingOptions = None,
        domain_endpoint_options: DomainEndpointOptions = None,
        advanced_security_options: AdvancedSecurityOptionsInput = None,
        node_to_node_encryption_options: NodeToNodeEncryptionOptions = None,
        encryption_at_rest_options: EncryptionAtRestOptions = None,
        auto_tune_options: AutoTuneOptions = None,
        dry_run: DryRun = None,
    ) -> UpdateElasticsearchDomainConfigResponse:
        """Modifies the cluster configuration of the specified Elasticsearch
        domain, setting as setting the instance type and the number of
        instances.

        :param domain_name: The name of the Elasticsearch domain that you are updating.
        :param elasticsearch_cluster_config: The type and number of instances to instantiate for the domain cluster.
        :param ebs_options: Specify the type and size of the EBS volume that you want to use.
        :param snapshot_options: Option to set the time, in UTC format, for the daily automated snapshot.
        :param vpc_options: Options to specify the subnets and security groups for VPC endpoint.
        :param cognito_options: Options to specify the Cognito user and identity pools for Kibana
        authentication.
        :param advanced_options: Modifies the advanced option to allow references to indices in an HTTP
        request body.
        :param access_policies: IAM access policy as a JSON-formatted string.
        :param log_publishing_options: Map of ``LogType`` and ``LogPublishingOption``, each containing options
        to publish a given type of Elasticsearch log.
        :param domain_endpoint_options: Options to specify configuration that will be applied to the domain
        endpoint.
        :param advanced_security_options: Specifies advanced security options.
        :param node_to_node_encryption_options: Specifies the NodeToNodeEncryptionOptions.
        :param encryption_at_rest_options: Specifies the Encryption At Rest Options.
        :param auto_tune_options: Specifies Auto-Tune options.
        :param dry_run: This flag, when set to True, specifies whether the
        ``UpdateElasticsearchDomain`` request should return the results of
        validation checks without actually applying the change.
        :returns: UpdateElasticsearchDomainConfigResponse
        :raises BaseException:
        :raises InternalException:
        :raises InvalidTypeException:
        :raises LimitExceededException:
        :raises ResourceNotFoundException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("UpdatePackage")
    def update_package(
        self,
        context: RequestContext,
        package_id: PackageID,
        package_source: PackageSource,
        package_description: PackageDescription = None,
        commit_message: CommitMessage = None,
    ) -> UpdatePackageResponse:
        """Updates a package for use with Amazon ES domains.

        :param package_id: Unique identifier for the package.
        :param package_source: The S3 location for importing the package specified as ``S3BucketName``
        and ``S3Key``.
        :param package_description: New description of the package.
        :param commit_message: An info message for the new version which will be shown as part of
        ``GetPackageVersionHistoryResponse``.
        :returns: UpdatePackageResponse
        :raises BaseException:
        :raises InternalException:
        :raises LimitExceededException:
        :raises ResourceNotFoundException:
        :raises AccessDeniedException:
        :raises ValidationException:
        """
        raise NotImplementedError

    @handler("UpgradeElasticsearchDomain")
    def upgrade_elasticsearch_domain(
        self,
        context: RequestContext,
        domain_name: DomainName,
        target_version: ElasticsearchVersionString,
        perform_check_only: Boolean = None,
    ) -> UpgradeElasticsearchDomainResponse:
        """Allows you to either upgrade your domain or perform an Upgrade
        eligibility check to a compatible Elasticsearch version.

        :param domain_name: The name of an Elasticsearch domain.
        :param target_version: The version of Elasticsearch that you intend to upgrade the domain to.
        :param perform_check_only: This flag, when set to True, indicates that an Upgrade Eligibility Check
        needs to be performed.
        :returns: UpgradeElasticsearchDomainResponse
        :raises BaseException:
        :raises ResourceNotFoundException:
        :raises ResourceAlreadyExistsException:
        :raises DisabledOperationException:
        :raises ValidationException:
        :raises InternalException:
        """
        raise NotImplementedError



import dataclasses
import io
import ipaddress
import logging
import os
import re
import shlex
import tarfile
import tempfile
from abc import ABCMeta, abstractmethod
from enum import Enum, unique
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

from localstack import config
from localstack.utils.common import TMP_FILES, HashableList, rm_rf, save_file, short_uid

LOG = logging.getLogger(__name__)


@unique
class DockerContainerStatus(Enum):
    DOWN = -1
    NON_EXISTENT = 0
    UP = 1
    PAUSED = 2


class ContainerException(Exception):
    def __init__(self, message=None, stdout=None, stderr=None) -> None:
        self.message = message or "Error during the communication with the docker daemon"
        self.stdout = stdout
        self.stderr = stderr


class NoSuchObject(ContainerException):
    def __init__(self, object_id: str, message=None, stdout=None, stderr=None) -> None:
        message = message or f"Docker object {object_id} not found"
        super().__init__(message, stdout, stderr)
        self.object_id = object_id


class NoSuchContainer(ContainerException):
    def __init__(self, container_name_or_id: str, message=None, stdout=None, stderr=None) -> None:
        message = message or f"Docker container {container_name_or_id} not found"
        super().__init__(message, stdout, stderr)
        self.container_name_or_id = container_name_or_id


class NoSuchImage(ContainerException):
    def __init__(self, image_name: str, message=None, stdout=None, stderr=None) -> None:
        message = message or f"Docker image {image_name} not found"
        super().__init__(message, stdout, stderr)
        self.image_name = image_name


class NoSuchNetwork(ContainerException):
    def __init__(self, network_name: str, message=None, stdout=None, stderr=None) -> None:
        message = message or f"Docker network {network_name} not found"
        super().__init__(message, stdout, stderr)
        self.network_name = network_name


class RegistryConnectionError(ContainerException):
    def __init__(self, details: str, message=None, stdout=None, stderr=None) -> None:
        message = message or f"Connection error: {details}"
        super().__init__(message, stdout, stderr)
        self.details = details


class AccessDenied(ContainerException):
    def __init__(self, object_name: str, message=None, stdout=None, stderr=None) -> None:
        message = message or f"Access denied to {object_name}"
        super().__init__(message, stdout, stderr)
        self.object_name = object_name


class PortMappings(object):
    """Maps source to target port ranges for Docker port mappings."""

    def __init__(self, bind_host=None):
        self.bind_host = bind_host if bind_host else ""
        self.mappings = {}

    def add(self, port, mapped=None, protocol="tcp"):
        mapped = mapped or port
        if isinstance(port, list):
            for i in range(port[1] - port[0] + 1):
                if isinstance(mapped, list):
                    self.add(port[0] + i, mapped[0] + i)
                else:
                    self.add(port[0] + i, mapped)
            return
        if port is None or int(port) <= 0:
            raise Exception("Unable to add mapping for invalid port: %s" % port)
        if self.contains(port):
            return
        bisected_host_port = None
        for from_range, to_range in self.mappings.items():
            if not self.in_expanded_range(port, from_range):
                continue
            if not self.in_expanded_range(mapped, to_range):
                continue
            from_range_len = from_range[1] - from_range[0]
            to_range_len = to_range[1] - to_range[0]
            is_uniform = from_range_len == to_range_len
            if is_uniform:
                self.expand_range(port, from_range)
                self.expand_range(mapped, to_range)
            else:
                if not self.in_range(mapped, to_range):
                    continue
                # extending a 1 to 1 mapping to be many to 1
                elif from_range_len == 1:
                    self.expand_range(port, from_range)
                # splitting a uniform mapping
                else:
                    bisected_port_index = mapped - to_range[0]
                    bisected_host_port = from_range[0] + bisected_port_index
                    self.bisect_range(mapped, to_range)
                    self.bisect_range(bisected_host_port, from_range)
                    break
            return
        protocol = str(protocol or "tcp").lower()
        if bisected_host_port is None:
            port_range = [port, port, protocol]
        elif bisected_host_port < port:
            port_range = [bisected_host_port, port, protocol]
        else:
            port_range = [port, bisected_host_port, protocol]
        self.mappings[HashableList(port_range)] = [mapped, mapped]

    def to_str(self) -> str:
        bind_address = f"{self.bind_host}:" if self.bind_host else ""

        def entry(k, v):
            protocol = "/%s" % k[2] if k[2] != "tcp" else ""
            if k[0] == k[1] and v[0] == v[1]:
                return "-p %s%s:%s%s" % (bind_address, k[0], v[0], protocol)
            if k[0] != k[1] and v[0] == v[1]:
                return "-p %s%s-%s:%s%s" % (bind_address, k[0], k[1], v[0], protocol)
            return "-p %s%s-%s:%s-%s%s" % (bind_address, k[0], k[1], v[0], v[1], protocol)

        return " ".join([entry(k, v) for k, v in self.mappings.items()])

    def to_list(self) -> List[str]:  # TODO test
        bind_address = f"{self.bind_host}:" if self.bind_host else ""

        def entry(k, v):
            protocol = "/%s" % k[2] if k[2] != "tcp" else ""
            if k[0] == k[1] and v[0] == v[1]:
                return ["-p", f"{bind_address}{k[0]}:{v[0]}{protocol}"]
            return ["-p", f"{bind_address}{k[0]}-{k[1]}:{v[0]}-{v[1]}{protocol}"]

        return [item for k, v in self.mappings.items() for item in entry(k, v)]

    def to_dict(self) -> Dict[str, Union[Tuple[str, Union[int, List[int]]], int]]:
        bind_address = self.bind_host or ""

        def entry(k, v):
            protocol = "/%s" % k[2]
            if k[0] != k[1] and v[0] == v[1]:
                container_port = v[0]
                host_ports = list(range(k[0], k[1] + 1))
                return [
                    (
                        f"{container_port}{protocol}",
                        (bind_address, host_ports) if bind_address else host_ports,
                    )
                ]
            return [
                (
                    f"{container_port}{protocol}",
                    (bind_address, host_port) if bind_address else host_port,
                )
                for container_port, host_port in zip(range(v[0], v[1] + 1), range(k[0], k[1] + 1))
            ]

        items = [item for k, v in self.mappings.items() for item in entry(k, v)]
        return dict(items)

    def contains(self, port):
        for from_range, to_range in self.mappings.items():
            if self.in_range(port, from_range):
                return True

    def in_range(self, port, range):
        return port >= range[0] and port <= range[1]

    def in_expanded_range(self, port, range):
        return port >= range[0] - 1 and port <= range[1] + 1

    def expand_range(self, port, range):
        if self.in_range(port, range):
            return
        if port == range[0] - 1:
            range[0] = port
        elif port == range[1] + 1:
            range[1] = port
        else:
            raise Exception("Unable to add port %s to existing range %s" % (port, range))

    """Bisect a port range, at the provided port
        This is needed in some cases when adding a non-uniform host to port mapping
        adjacent to an existing port range
    """

    def bisect_range(self, port, range):
        if not self.in_range(port, range):
            return
        if port == range[0]:
            range[0] = port + 1
        else:
            range[1] = port - 1

    def __repr__(self):
        return f"<PortMappings: {self.to_dict()}>"


SimpleVolumeBind = Tuple[str, str]


@dataclasses.dataclass
class VolumeBind:
    """Represents a --volume argument run/create command. When using VolumeBind to bind-mount a file or directory
    that does not yet exist on the Docker host, -v creates the endpoint for you. It is always created as a directory.
    """

    host_dir: str
    container_dir: str
    options: Optional[List[str]] = None

    def to_str(self) -> str:
        args = []

        if self.host_dir:
            args.append(self.host_dir)

        if not self.container_dir:
            raise ValueError("no container dir specified")

        args.append(self.container_dir)

        if self.options:
            args.append(self.options)

        return ":".join(args)


class VolumeMappings:
    mappings: List[Union[SimpleVolumeBind, VolumeBind]]

    def __init__(self, mappings: List[Union[SimpleVolumeBind, VolumeBind]] = None):
        self.mappings = mappings if mappings is not None else []

    def add(self, mapping: Union[SimpleVolumeBind, VolumeBind]):
        self.append(mapping)

    def append(
        self,
        mapping: Union[
            SimpleVolumeBind,
            VolumeBind,
        ],
    ):
        self.mappings.append(mapping)

    def __iter__(self):
        return self.mappings.__iter__()


@dataclasses.dataclass
class ContainerConfiguration:
    image_name: str
    name: Optional[str] = None
    volumes: Optional[VolumeMappings] = None
    ports: Optional[PortMappings] = None
    entrypoint: Optional[str] = None
    additional_flags: Optional[List[str]] = None
    command: Optional[List[str]] = None
    env_vars: Dict[str, str] = dataclasses.field(default_factory=dict)

    privileged: Optional[bool] = None
    remove: Optional[bool] = None
    interactive: Optional[bool] = None
    tty: Optional[bool] = None
    detach: Optional[bool] = None

    stdin: Optional[str] = None
    user: Optional[str] = None
    cap_add: Optional[str] = None
    network: Optional[str] = None
    dns: Optional[str] = None
    workdir: Optional[str] = None


class ContainerClient(metaclass=ABCMeta):
    STOP_TIMEOUT = 0

    @abstractmethod
    def get_container_status(self, container_name: str) -> DockerContainerStatus:
        """Returns the status of the container with the given name"""
        pass

    def get_networks(self, container_name: str) -> List[str]:
        LOG.debug("Getting networks for container: %s", container_name)
        container_attrs = self.inspect_container(container_name_or_id=container_name)
        return list(container_attrs["NetworkSettings"]["Networks"].keys())

    def get_container_ipv4_for_network(
        self, container_name_or_id: str, container_network: str
    ) -> str:
        """
        Returns the IPv4 address for the container on the interface connected to the given network
        :param container_name_or_id: Container to inspect
        :param container_network: Network the IP address will belong to
        :return: IP address of the given container on the interface connected to the given network
        """
        LOG.debug(
            "Getting ipv4 address for container %s in network %s.",
            container_name_or_id,
            container_network,
        )
        # we always need the ID for this
        container_id = self.get_container_id(container_name=container_name_or_id)
        network_attrs = self.inspect_network(container_network)
        containers = network_attrs["Containers"]
        if container_id not in containers:
            raise ContainerException(
                "Container %s is not connected to target network %s",
                container_name_or_id,
                container_network,
            )
        try:
            ip = str(ipaddress.IPv4Interface(containers[container_id]["IPv4Address"]).ip)
        except Exception as e:
            raise ContainerException(
                f"Unable to detect IP address for container {container_name_or_id} in network {container_network}: {e}"
            )
        return ip

    @abstractmethod
    def stop_container(self, container_name: str, timeout: int = None):
        """Stops container with given name
        :param container_name: Container identifier (name or id) of the container to be stopped
        :param timeout: Timeout after which SIGKILL is sent to the container.
                        If not specified, defaults to `STOP_TIMEOUT`
        """
        pass

    @abstractmethod
    def pause_container(self, container_name: str):
        """Pauses a container with the given name."""

    @abstractmethod
    def unpause_container(self, container_name: str):
        """Unpauses a container with the given name."""

    @abstractmethod
    def remove_container(self, container_name: str, force=True, check_existence=False) -> None:
        """Removes container with given name"""
        pass

    @abstractmethod
    def remove_image(self, image: str, force: bool = True) -> None:
        """Removes an image with given name

        :param image: Image name and tag
        :param force: Force removal
        """
        pass

    @abstractmethod
    def list_containers(self, filter: Union[List[str], str, None] = None, all=True) -> List[dict]:
        """List all containers matching the given filters

        :return: A list of dicts with keys id, image, name, labels, status
        """
        pass

    def get_running_container_names(self) -> List[str]:
        """Returns a list of the names of all running containers"""
        result = self.list_containers(all=False)
        result = list(map(lambda container: container["name"], result))
        return result

    def is_container_running(self, container_name: str) -> bool:
        """Checks whether a container with a given name is currently running"""
        return container_name in self.get_running_container_names()

    @abstractmethod
    def copy_into_container(
        self, container_name: str, local_path: str, container_path: str
    ) -> None:
        """Copy contents of the given local path into the container"""
        pass

    @abstractmethod
    def copy_from_container(
        self, container_name: str, local_path: str, container_path: str
    ) -> None:
        """Copy contents of the given container to the host"""
        pass

    @abstractmethod
    def pull_image(self, docker_image: str) -> None:
        """Pulls a image with a given name from a Docker registry"""
        pass

    @abstractmethod
    def push_image(self, docker_image: str) -> None:
        """Pushes a image with a given name to a Docker registry"""
        pass

    @abstractmethod
    def build_image(self, dockerfile_path: str, image_name: str, context_path: str = None) -> None:
        """Builds an image from the given Dockerfile

        :param dockerfile_path: Path to Dockerfile, or a directory that contains a Dockerfile
        :param image_name: Name of the image to be built
        :param context_path: Path for build context (defaults to dirname of Dockerfile)
        """
        pass

    @abstractmethod
    def tag_image(self, source_ref: str, target_name: str) -> None:
        """Tags an image with a new name

        :param source_ref: Name or ID of the image to be tagged
        :param target_name: New name (tag) of the tagged image
        """
        pass

    @abstractmethod
    def get_docker_image_names(self, strip_latest=True, include_tags=True) -> List[str]:
        """
        Get all names of docker images available to the container engine
        :param strip_latest: return images both with and without :latest tag
        :param include_tags: Include tags of the images in the names
        :return: List of image names
        """
        pass

    @abstractmethod
    def get_container_logs(self, container_name_or_id: str, safe=False) -> str:
        """Get all logs of a given container"""
        pass

    @abstractmethod
    def inspect_container(self, container_name_or_id: str) -> Dict[str, Union[Dict, str]]:
        """Get detailed attributes of an container.

        :return: Dict containing docker attributes as returned by the daemon
        """
        pass

    @abstractmethod
    def inspect_image(self, image_name: str, pull: bool = True) -> Dict[str, Union[Dict, str]]:
        """Get detailed attributes of an image.

        :param image_name: Image name to inspect
        :param pull: Whether to pull image if not existent
        :return: Dict containing docker attributes as returned by the daemon
        """
        pass

    @abstractmethod
    def inspect_network(self, network_name: str) -> Dict[str, Union[Dict, str]]:
        """Get detailed attributes of an network.

        :return: Dict containing docker attributes as returned by the daemon
        """
        pass

    @abstractmethod
    def connect_container_to_network(
        self, network_name: str, container_name_or_id: str, aliases: Optional[List] = None
    ) -> None:
        """
        Connects a container to a given network
        :param network_name: Network to connect the container to
        :param container_name_or_id: Container to connect to the network
        :param aliases: List of dns names the container should be available under in the network
        """
        pass

    @abstractmethod
    def disconnect_container_from_network(
        self, network_name: str, container_name_or_id: str
    ) -> None:
        """
        Disconnects a container from a given network
        :param network_name: Network to disconnect the container from
        :param container_name_or_id: Container to disconnect from the network
        """
        pass

    def get_container_name(self, container_id: str) -> str:
        """Get the name of a container by a given identifier"""
        return self.inspect_container(container_id)["Name"].lstrip("/")

    def get_container_id(self, container_name: str) -> str:
        """Get the id of a container by a given name"""
        return self.inspect_container(container_name)["Id"]

    @abstractmethod
    def get_container_ip(self, container_name_or_id: str) -> str:
        """Get the IP address of a given container

        If container has multiple networks, it will return the IP of the first
        """
        pass

    def get_image_cmd(self, docker_image: str, pull: bool = True) -> List[str]:
        """Get the command for the given image
        :param docker_image: Docker image to inspect
        :param pull: Whether to pull if image is not present
        :return: Image command in its array form
        """
        cmd_list = self.inspect_image(docker_image, pull)["Config"]["Cmd"] or []
        return cmd_list

    def get_image_entrypoint(self, docker_image: str, pull: bool = True) -> str:
        """Get the entry point for the given image
        :param docker_image: Docker image to inspect
        :param pull: Whether to pull if image is not present
        :return: Image entrypoint
        """
        LOG.debug("Getting the entrypoint for image: %s", docker_image)
        entrypoint_list = self.inspect_image(docker_image, pull)["Config"]["Entrypoint"] or []
        return shlex.join(entrypoint_list)

    @abstractmethod
    def has_docker(self) -> bool:
        """Check if system has docker available"""
        pass

    @abstractmethod
    def commit(
        self,
        container_name_or_id: str,
        image_name: str,
        image_tag: str,
    ):
        """Create an image from a running container.

        :param container_name_or_id: Source container
        :param image_name: Destination image name
        :param image_tag: Destination image tag
        """
        pass

    @abstractmethod
    def create_container(
        self,
        image_name: str,
        *,
        name: Optional[str] = None,
        entrypoint: Optional[str] = None,
        remove: bool = False,
        interactive: bool = False,
        tty: bool = False,
        detach: bool = False,
        command: Optional[Union[List[str], str]] = None,
        mount_volumes: Optional[List[SimpleVolumeBind]] = None,
        ports: Optional[PortMappings] = None,
        env_vars: Optional[Dict[str, str]] = None,
        user: Optional[str] = None,
        cap_add: Optional[str] = None,
        network: Optional[str] = None,
        dns: Optional[str] = None,
        additional_flags: Optional[str] = None,
        workdir: Optional[str] = None,
    ) -> str:
        """Creates a container with the given image

        :return: Container ID
        """
        pass

    @abstractmethod
    def run_container(
        self,
        image_name: str,
        stdin: bytes = None,
        *,
        name: Optional[str] = None,
        entrypoint: Optional[str] = None,
        remove: bool = False,
        interactive: bool = False,
        tty: bool = False,
        detach: bool = False,
        command: Optional[Union[List[str], str]] = None,
        mount_volumes: Optional[List[SimpleVolumeBind]] = None,
        ports: Optional[PortMappings] = None,
        env_vars: Optional[Dict[str, str]] = None,
        user: Optional[str] = None,
        cap_add: Optional[str] = None,
        network: Optional[str] = None,
        dns: Optional[str] = None,
        additional_flags: Optional[str] = None,
        workdir: Optional[str] = None,
    ) -> Tuple[bytes, bytes]:
        """Creates and runs a given docker container

        :return: A tuple (stdout, stderr)
        """
        pass

    @abstractmethod
    def exec_in_container(
        self,
        container_name_or_id: str,
        command: Union[List[str], str],
        interactive: bool = False,
        detach: bool = False,
        env_vars: Optional[Dict[str, Optional[str]]] = None,
        stdin: Optional[bytes] = None,
        user: Optional[str] = None,
    ) -> Tuple[bytes, bytes]:
        """Execute a given command in a container

        :return: A tuple (stdout, stderr)
        """
        pass

    @abstractmethod
    def start_container(
        self,
        container_name_or_id: str,
        stdin: bytes = None,
        interactive: bool = False,
        attach: bool = False,
        flags: Optional[str] = None,
    ) -> Tuple[bytes, bytes]:
        """Start a given, already created container

        :return: A tuple (stdout, stderr) if attach or interactive is set, otherwise a tuple (b"container_name_or_id", b"")
        """
        pass


class Util:
    MAX_ENV_ARGS_LENGTH = 20000

    @staticmethod
    def format_env_vars(key: str, value: Optional[str]):
        if value is None:
            return key
        return f"{key}={value}"

    @classmethod
    def create_env_vars_file_flag(cls, env_vars: Dict) -> Tuple[List[str], Optional[str]]:
        if not env_vars:
            return [], None
        result = []
        env_vars = dict(env_vars)
        env_file = None
        if len(str(env_vars)) > cls.MAX_ENV_ARGS_LENGTH:
            # default ARG_MAX=131072 in Docker - let's create an env var file if the string becomes too long...
            env_file = cls.mountable_tmp_file()
            env_content = ""
            for name, value in dict(env_vars).items():
                if len(value) > cls.MAX_ENV_ARGS_LENGTH:
                    # each line in the env file has a max size as well (error "bufio.Scanner: token too long")
                    continue
                env_vars.pop(name)
                value = value.replace("\n", "\\")
                env_content += f"{cls.format_env_vars(name, value)}\n"
            save_file(env_file, env_content)
            result += ["--env-file", env_file]

        env_vars_res = [
            item for k, v in env_vars.items() for item in ["-e", cls.format_env_vars(k, v)]
        ]
        result += env_vars_res
        return result, env_file

    @staticmethod
    def rm_env_vars_file(env_vars_file) -> None:
        if env_vars_file:
            return rm_rf(env_vars_file)

    @staticmethod
    def mountable_tmp_file():
        f = os.path.join(config.dirs.tmp, short_uid())
        TMP_FILES.append(f)
        return f

    @staticmethod
    def append_without_latest(image_names):
        suffix = ":latest"
        for image in list(image_names):
            if image.endswith(suffix):
                image_names.append(image[: -len(suffix)])

    @staticmethod
    def tar_path(path, target_path, is_dir: bool):
        f = tempfile.NamedTemporaryFile()
        with tarfile.open(mode="w", fileobj=f) as t:
            abs_path = os.path.abspath(path)
            arcname = (
                os.path.basename(path)
                if is_dir
                else (os.path.basename(target_path) or os.path.basename(path))
            )
            t.add(abs_path, arcname=arcname)

        f.seek(0)
        return f

    @staticmethod
    def untar_to_path(tardata, target_path):
        target_path = Path(target_path)
        with tarfile.open(mode="r", fileobj=io.BytesIO(b"".join(b for b in tardata))) as t:
            if target_path.is_dir():
                t.extractall(path=target_path)
            else:
                member = t.next()
                if member:
                    member.name = target_path.name
                    t.extract(member, target_path.parent)
                else:
                    LOG.debug("File to copy empty, ignoring...")

    @staticmethod
    def parse_additional_flags(
        additional_flags: str,
        env_vars: Dict[str, str] = None,
        ports: PortMappings = None,
        mounts: List[SimpleVolumeBind] = None,
        network: Optional[str] = None,
    ) -> Tuple[
        Dict[str, str],
        PortMappings,
        List[SimpleVolumeBind],
        Optional[Dict[str, str]],
        Optional[str],
    ]:
        """Parses environment, volume and port flags passed as string
        :param additional_flags: String which contains the flag definitions
        :param env_vars: Dict with env vars. Will be modified in place.
        :param ports: PortMapping object. Will be modified in place.
        :param mounts: List of mount tuples (host_path, container_path). Will be modified in place.
        :param network: Existing network name (optional). Warning will be printed if network is overwritten in flags.
        :return: A tuple containing the env_vars, ports, mount, extra_hosts and network objects. Will return new objects
                if respective parameters were None and additional flags contained a flag for that object, the same which
                are passed otherwise.
        """
        cur_state = None
        extra_hosts = None
        # TODO Use argparse to simplify this logic
        for flag in shlex.split(additional_flags):
            if not cur_state:
                if flag in ["-v", "--volume"]:
                    cur_state = "volume"
                elif flag in ["-p", "--publish"]:
                    cur_state = "port"
                elif flag in ["-e", "--env"]:
                    cur_state = "env"
                elif flag == "--add-host":
                    cur_state = "add-host"
                elif flag == "--network":
                    cur_state = "set-network"
                else:
                    raise NotImplementedError(
                        f"Flag {flag} is currently not supported by this Docker client."
                    )
            else:
                if cur_state == "volume":
                    mounts = mounts if mounts is not None else []
                    match = re.match(
                        r"(?P<host>[\w\s\\\/:\-.]+?):(?P<container>[\w\s\/\-.]+)(?::(?P<arg>ro|rw|z|Z))?",
                        flag,
                    )
                    if not match:
                        LOG.warning("Unable to parse volume mount Docker flags: %s", flag)
                        continue
                    host_path = match.group("host")
                    container_path = match.group("container")
                    rw_args = match.group("arg")
                    if rw_args:
                        LOG.info("Volume options like :ro or :rw are currently ignored.")
                    mounts.append((host_path, container_path))
                elif cur_state == "port":
                    port_split = flag.split(":")
                    protocol = "tcp"
                    if len(port_split) == 2:
                        host_port, container_port = port_split
                    elif len(port_split) == 3:
                        LOG.warning(
                            "Host part of port mappings are ignored currently in additional flags"
                        )
                        _, host_port, container_port = port_split
                    else:
                        raise ValueError("Invalid port string provided: %s", flag)
                    host_port_split = host_port.split("-")
                    if len(host_port_split) == 2:
                        host_port = [int(host_port_split[0]), int(host_port_split[1])]
                    elif len(host_port_split) == 1:
                        host_port = int(host_port)
                    else:
                        raise ValueError("Invalid port string provided: %s", flag)
                    if "/" in container_port:
                        container_port, protocol = container_port.split("/")
                    ports = ports if ports is not None else PortMappings()
                    ports.add(host_port, int(container_port), protocol)
                elif cur_state == "env":
                    lhs, _, rhs = flag.partition("=")
                    env_vars = env_vars if env_vars is not None else {}
                    env_vars[lhs] = rhs
                elif cur_state == "add-host":
                    extra_hosts = extra_hosts if extra_hosts is not None else {}
                    hosts_split = flag.split(":")
                    extra_hosts[hosts_split[0]] = hosts_split[1]
                elif cur_state == "set-network":
                    if network:
                        LOG.warning(
                            "Overwriting Docker container network '%s' with new value '%s'",
                            network,
                            flag,
                        )
                    network = flag

                cur_state = None
        return env_vars, ports, mounts, extra_hosts, network

    @staticmethod
    def convert_mount_list_to_dict(
        mount_volumes: List[SimpleVolumeBind],
    ) -> Dict[str, Dict[str, str]]:
        """Converts a List of (host_path, container_path) tuples to a Dict suitable as volume argument for docker sdk"""
        return dict(
            map(
                lambda paths: (str(paths[0]), {"bind": paths[1], "mode": "rw"}),
                mount_volumes,
            )
        )

    @staticmethod
    def resolve_dockerfile_path(dockerfile_path: str) -> str:
        """If the given path is a directory that contains a Dockerfile, then return the file path to it."""
        rel_path = os.path.join(dockerfile_path, "Dockerfile")
        if os.path.isdir(dockerfile_path) and os.path.exists(rel_path):
            return rel_path
        return dockerfile_path



from urllib.parse import parse_qs, urlparse

from moto.route53 import responses as route53_responses
from moto.route53.models import route53_backend
from moto.route53.responses import DELETE_HEALTH_CHECK_RESPONSE
from moto.route53.urls import url_paths as route53_url_paths

from localstack import config
from localstack.services.infra import start_moto_server
from localstack.utils.patch import patch

XMLNS_ROUTE53 = "https://route53.amazonaws.com/doc/2013-04-01/"


def apply_patches():
    # patch function to match hosted zone if zone name ends with a dot
    @patch(route53_responses.Route53.list_hosted_zones_by_name_response)
    def list_hosted_zones_by_name_response(fn, self, request, full_url, headers):
        parsed_url = urlparse(full_url)
        query_params = parse_qs(parsed_url.query)
        dnsname = query_params.get("dnsname")
        dnsname = dnsname and dnsname[0]
        all_zones = route53_backend.get_all_hosted_zones()
        zones1 = [zone for zone in all_zones if zone.name == dnsname]
        zones2 = [zone for zone in all_zones if zone.name == "%s." % dnsname]
        if not zones1 and zones2:
            full_url = full_url.replace("dnsname=%s" % dnsname, "dnsname=%s." % dnsname)
        return fn(self, request, full_url, headers)

    def get_or_delete_health_check(self, request, full_url, headers):
        parsed_url = urlparse(full_url)
        parts = parsed_url.path.strip("/").split("/")
        health_check_id = parts[-1]
        if request.method == "GET":
            health_check = route53_backend.health_checks.get(health_check_id)
            if not health_check:
                return 404, {}, ""
            result = """<GetHealthCheckResponse xmlns="{}">{}</GetHealthCheckResponse>"""
            result = result.format(XMLNS_ROUTE53, health_check.to_xml()).strip()
            return result
        if request.method == "DELETE":
            route53_backend.delete_health_check(health_check_id)
            return 200, headers, DELETE_HEALTH_CHECK_RESPONSE

    if not hasattr(route53_responses.Route53, "get_or_delete_health_check"):
        route53_responses.Route53.get_or_delete_health_check = get_or_delete_health_check

    # update URL path mappings to enable the patch
    path_regex = r"{0}/(?P<api_version>[\d_-]+)/healthcheck/(?P<health_check_id>[^/]+)/?$"
    route53_url_paths[path_regex] = route53_responses.Route53().get_or_delete_health_check


def start_route53(port=None, asynchronous=False, update_listener=None):
    port = port or config.service_port("route53")
    apply_patches()
    return start_moto_server(
        "route53",
        port,
        name="Route53",
        asynchronous=asynchronous,
        update_listener=update_listener,
    )


def start_route53_resolver(port=None, asynchronous=False, update_listener=None):
    port = port or config.service_port("route53resolver")
    return start_moto_server(
        "route53resolver",
        port,
        name="Route53 Resolver",
        asynchronous=asynchronous,
        update_listener=update_listener,
    )



from contextlib import contextmanager
from typing import Optional, cast

from botocore.exceptions import ClientError

from localstack import constants
from localstack.aws.api import RequestContext
from localstack.aws.api.es import (
    ARN,
    AccessDeniedException,
    AdvancedOptions,
    AdvancedSecurityOptionsInput,
    AutoTuneOptionsInput,
)
from localstack.aws.api.es import BaseException as EsBaseException
from localstack.aws.api.es import (
    CognitoOptions,
    CompatibleElasticsearchVersionsList,
    CompatibleVersionsMap,
    ConflictException,
    CreateElasticsearchDomainResponse,
    DeleteElasticsearchDomainResponse,
    DescribeElasticsearchDomainConfigResponse,
    DescribeElasticsearchDomainResponse,
    DescribeElasticsearchDomainsResponse,
    DisabledOperationException,
    DomainEndpointOptions,
    DomainInfoList,
    DomainName,
    DomainNameList,
    EBSOptions,
    ElasticsearchClusterConfig,
    ElasticsearchClusterConfigStatus,
    ElasticsearchDomainConfig,
    ElasticsearchDomainStatus,
    ElasticsearchVersionStatus,
    ElasticsearchVersionString,
    EncryptionAtRestOptions,
    EngineType,
    EsApi,
    GetCompatibleElasticsearchVersionsResponse,
    InternalException,
    InvalidPaginationTokenException,
    InvalidTypeException,
    LimitExceededException,
    ListDomainNamesResponse,
    ListElasticsearchVersionsResponse,
    ListTagsResponse,
    LogPublishingOptions,
    MaxResults,
    NextToken,
    NodeToNodeEncryptionOptions,
    OptionStatus,
    PolicyDocument,
    ResourceAlreadyExistsException,
    ResourceNotFoundException,
    SnapshotOptions,
    StringList,
    TagList,
    ValidationException,
    VPCOptions,
)
from localstack.aws.api.opensearch import (
    ClusterConfig,
    CompatibleVersionsList,
    DomainConfig,
    DomainStatus,
    VersionString,
)
from localstack.utils.analytics import event_publisher
from localstack.utils.aws import aws_stack


def _version_to_opensearch(
    version: Optional[ElasticsearchVersionString],
) -> Optional[VersionString]:
    if version is not None:
        if version.startswith("OpenSearch_"):
            return version
        else:
            return f"Elasticsearch_{version}"


def _version_from_opensearch(
    version: Optional[VersionString],
) -> Optional[ElasticsearchVersionString]:
    if version is not None:
        if version.startswith("Elasticsearch_"):
            return version.split("_")[1]
        else:
            return version


def _instancetype_to_opensearch(instance_type: Optional[str]) -> Optional[str]:
    if instance_type is not None:
        return instance_type.replace("elasticsearch", "search")


def _instancetype_from_opensearch(instance_type: Optional[str]) -> Optional[str]:
    if instance_type is not None:
        return instance_type.replace("search", "elasticsearch")


def _clusterconfig_from_opensearch(
    cluster_config: Optional[ClusterConfig],
) -> Optional[ElasticsearchClusterConfig]:
    if cluster_config is not None:
        # Just take the whole typed dict and typecast it to our target type
        result = cast(ElasticsearchClusterConfig, cluster_config)

        # Adjust the instance type names
        result["InstanceType"] = _instancetype_from_opensearch(cluster_config.get("InstanceType"))
        result["DedicatedMasterType"] = _instancetype_from_opensearch(
            cluster_config.get("DedicatedMasterType")
        )
        result["WarmType"] = _instancetype_from_opensearch(cluster_config.get("WarmType"))
        return result


def _domainstatus_from_opensearch(
    domain_status: Optional[DomainStatus],
) -> Optional[ElasticsearchDomainStatus]:
    if domain_status is not None:
        # Just take the whole typed dict and typecast it to our target type
        result = cast(ElasticsearchDomainStatus, domain_status)
        # Only specifically handle keys which are named differently or their values differ (version and clusterconfig)
        result["ElasticsearchVersion"] = _version_from_opensearch(
            domain_status.get("EngineVersion")
        )
        result["ElasticsearchClusterConfig"] = _clusterconfig_from_opensearch(
            domain_status.get("ClusterConfig")
        )
        result.pop("EngineVersion", None)
        result.pop("ClusterConfig", None)
        return result


def _clusterconfig_to_opensearch(
    elasticsearch_cluster_config: Optional[ElasticsearchClusterConfig],
) -> Optional[ClusterConfig]:
    if elasticsearch_cluster_config is not None:
        result = cast(ClusterConfig, elasticsearch_cluster_config)
        result["InstanceType"] = _instancetype_to_opensearch(result.get("InstanceType"))
        result["DedicatedMasterType"] = _instancetype_to_opensearch(
            result.get("DedicatedMasterType")
        )
        result["WarmType"] = _instancetype_to_opensearch(result.get("WarmType"))
        return result


def _domainconfig_from_opensearch(
    domain_config: Optional[DomainConfig],
) -> Optional[ElasticsearchDomainConfig]:
    if domain_config is not None:
        result = cast(ElasticsearchDomainConfig, domain_config)
        engine_version = domain_config.get("EngineVersion", {})
        result["ElasticsearchVersion"] = ElasticsearchVersionStatus(
            Options=_version_from_opensearch(engine_version.get("Options")),
            Status=cast(OptionStatus, engine_version.get("Status")),
        )
        cluster_config = domain_config.get("ClusterConfig", {})
        result["ElasticsearchClusterConfig"] = ElasticsearchClusterConfigStatus(
            Options=_clusterconfig_from_opensearch(cluster_config.get("Options")),
            Status=cluster_config.get("Status"),
        )
        result.pop("EngineVersion", None)
        result.pop("ClusterConfig", None)
        return result


def _compatible_version_list_from_opensearch(
    compatible_version_list: Optional[CompatibleVersionsList],
) -> Optional[CompatibleElasticsearchVersionsList]:
    if compatible_version_list is not None:
        return [
            CompatibleVersionsMap(
                SourceVersion=_version_from_opensearch(version_map["SourceVersion"]),
                TargetVersions=[
                    _version_from_opensearch(target_version)
                    for target_version in version_map["TargetVersions"]
                ],
            )
            for version_map in compatible_version_list
        ]


@contextmanager
def exception_mapper():
    """Maps an exception thrown by the OpenSearch client to an exception thrown by the ElasticSearch API."""
    try:
        yield
    except ClientError as err:
        exception_types = {
            "AccessDeniedException": AccessDeniedException,
            "BaseException": EsBaseException,
            "ConflictException": ConflictException,
            "DisabledOperationException": DisabledOperationException,
            "InternalException": InternalException,
            "InvalidPaginationTokenException": InvalidPaginationTokenException,
            "InvalidTypeException": InvalidTypeException,
            "LimitExceededException": LimitExceededException,
            "ResourceAlreadyExistsException": ResourceAlreadyExistsException,
            "ResourceNotFoundException": ResourceNotFoundException,
            "ValidationException": ValidationException,
        }
        mapped_exception_type = exception_types.get(err.response["Error"]["Code"], EsBaseException)
        raise mapped_exception_type(err.response["Error"]["Message"])


class EsProvider(EsApi):
    def create_elasticsearch_domain(
        self,
        context: RequestContext,
        domain_name: DomainName,
        elasticsearch_version: ElasticsearchVersionString = None,
        elasticsearch_cluster_config: ElasticsearchClusterConfig = None,
        ebs_options: EBSOptions = None,
        access_policies: PolicyDocument = None,
        snapshot_options: SnapshotOptions = None,
        vpc_options: VPCOptions = None,
        cognito_options: CognitoOptions = None,
        encryption_at_rest_options: EncryptionAtRestOptions = None,
        node_to_node_encryption_options: NodeToNodeEncryptionOptions = None,
        advanced_options: AdvancedOptions = None,
        log_publishing_options: LogPublishingOptions = None,
        domain_endpoint_options: DomainEndpointOptions = None,
        advanced_security_options: AdvancedSecurityOptionsInput = None,
        auto_tune_options: AutoTuneOptionsInput = None,
        tag_list: TagList = None,
    ) -> CreateElasticsearchDomainResponse:
        opensearch_client = aws_stack.connect_to_service("opensearch", region_name=context.region)
        # If no version is given, we set our default elasticsearch version
        engine_version = (
            _version_to_opensearch(elasticsearch_version)
            if elasticsearch_version
            else constants.ELASTICSEARCH_DEFAULT_VERSION
        )
        kwargs = {
            "DomainName": domain_name,
            "EngineVersion": engine_version,
            "ClusterConfig": _clusterconfig_to_opensearch(elasticsearch_cluster_config),
            "EBSOptions": ebs_options,
            "AccessPolicies": access_policies,
            "SnapshotOptions": snapshot_options,
            "VPCOptions": vpc_options,
            "CognitoOptions": cognito_options,
            "EncryptionAtRestOptions": encryption_at_rest_options,
            "NodeToNodeEncryptionOptions": node_to_node_encryption_options,
            "AdvancedOptions": advanced_options,
            "LogPublishingOptions": log_publishing_options,
            "DomainEndpointOptions": domain_endpoint_options,
            "AdvancedSecurityOptions": advanced_security_options,
            "AutoTuneOptions": auto_tune_options,
            "TagList": tag_list,
        }

        # Filter the kwargs to not set None values at all (boto doesn't like that)
        kwargs = {key: value for key, value in kwargs.items() if value is not None}

        with exception_mapper():
            domain_status = opensearch_client.create_domain(**kwargs)["DomainStatus"]

        # record event
        event_publisher.fire_event(
            event_publisher.EVENT_ES_CREATE_DOMAIN,
            payload={"n": event_publisher.get_hash(domain_name)},
        )

        status = _domainstatus_from_opensearch(domain_status)
        return CreateElasticsearchDomainResponse(DomainStatus=status)

    def delete_elasticsearch_domain(
        self, context: RequestContext, domain_name: DomainName
    ) -> DeleteElasticsearchDomainResponse:
        opensearch_client = aws_stack.connect_to_service("opensearch", region_name=context.region)

        with exception_mapper():
            domain_status = opensearch_client.delete_domain(
                DomainName=domain_name,
            )["DomainStatus"]

        # record event
        event_publisher.fire_event(
            event_publisher.EVENT_ES_DELETE_DOMAIN,
            payload={"n": event_publisher.get_hash(domain_name)},
        )

        status = _domainstatus_from_opensearch(domain_status)
        return DeleteElasticsearchDomainResponse(DomainStatus=status)

    def describe_elasticsearch_domain(
        self, context: RequestContext, domain_name: DomainName
    ) -> DescribeElasticsearchDomainResponse:
        opensearch_client = aws_stack.connect_to_service("opensearch", region_name=context.region)

        with exception_mapper():
            opensearch_status = opensearch_client.describe_domain(
                DomainName=domain_name,
            )["DomainStatus"]

        status = _domainstatus_from_opensearch(opensearch_status)
        return DescribeElasticsearchDomainResponse(DomainStatus=status)

    def describe_elasticsearch_domains(
        self, context: RequestContext, domain_names: DomainNameList
    ) -> DescribeElasticsearchDomainsResponse:
        opensearch_client = aws_stack.connect_to_service("opensearch", region_name=context.region)

        with exception_mapper():
            opensearch_status_list = opensearch_client.describe_domains(
                DomainNames=domain_names,
            )["DomainStatusList"]

        status_list = [_domainstatus_from_opensearch(s) for s in opensearch_status_list]
        return DescribeElasticsearchDomainsResponse(DomainStatusList=status_list)

    def list_domain_names(
        self, context: RequestContext, engine_type: EngineType = None
    ) -> ListDomainNamesResponse:
        opensearch_client = aws_stack.connect_to_service("opensearch", region_name=context.region)
        # Only hand the EngineType param to boto if it's set
        kwargs = {}
        if engine_type:
            kwargs["EngineType"] = engine_type

        with exception_mapper():
            domain_names = opensearch_client.list_domain_names(**kwargs)["DomainNames"]

        return ListDomainNamesResponse(DomainNames=cast(Optional[DomainInfoList], domain_names))

    def list_elasticsearch_versions(
        self,
        context: RequestContext,
        max_results: MaxResults = None,
        next_token: NextToken = None,
    ) -> ListElasticsearchVersionsResponse:
        opensearch_client = aws_stack.connect_to_service("opensearch", region_name=context.region)
        # Construct the arguments as kwargs to not set None values at all (boto doesn't like that)
        kwargs = {
            key: value
            for key, value in {"MaxResults": max_results, "NextToken": next_token}.items()
            if value is not None
        }
        with exception_mapper():
            versions = opensearch_client.list_versions(**kwargs)

        return ListElasticsearchVersionsResponse(
            ElasticsearchVersions=[
                _version_from_opensearch(version) for version in versions["Versions"]
            ],
            NextToken=versions.get(next_token),
        )

    def get_compatible_elasticsearch_versions(
        self, context: RequestContext, domain_name: DomainName = None
    ) -> GetCompatibleElasticsearchVersionsResponse:
        opensearch_client = aws_stack.connect_to_service("opensearch", region_name=context.region)
        # Only hand the DomainName param to boto if it's set
        kwargs = {}
        if domain_name:
            kwargs["DomainName"] = domain_name

        with exception_mapper():
            compatible_versions_response = opensearch_client.get_compatible_versions(**kwargs)

        compatible_versions = compatible_versions_response.get("CompatibleVersions")
        return GetCompatibleElasticsearchVersionsResponse(
            CompatibleElasticsearchVersions=_compatible_version_list_from_opensearch(
                compatible_versions
            )
        )

    def describe_elasticsearch_domain_config(
        self, context: RequestContext, domain_name: DomainName
    ) -> DescribeElasticsearchDomainConfigResponse:
        opensearch_client = aws_stack.connect_to_service("opensearch", region_name=context.region)

        with exception_mapper():
            domain_config = opensearch_client.describe_domain_config(DomainName=domain_name).get(
                "DomainConfig"
            )

        return DescribeElasticsearchDomainConfigResponse(
            DomainConfig=_domainconfig_from_opensearch(domain_config)
        )

    def add_tags(self, context: RequestContext, arn: ARN, tag_list: TagList) -> None:
        opensearch_client = aws_stack.connect_to_service("opensearch", region_name=context.region)

        with exception_mapper():
            opensearch_client.add_tags(ARN=arn, TagList=tag_list)

    def list_tags(self, context: RequestContext, arn: ARN) -> ListTagsResponse:
        opensearch_client = aws_stack.connect_to_service("opensearch", region_name=context.region)

        with exception_mapper():
            response = opensearch_client.list_tags(ARN=arn)

        return ListTagsResponse(TagList=response.get("TagList"))

    def remove_tags(self, context: RequestContext, arn: ARN, tag_keys: StringList) -> None:
        opensearch_client = aws_stack.connect_to_service("opensearch", region_name=context.region)

        with exception_mapper():
            opensearch_client.remove_tags(ARN=arn, TagKeys=tag_keys)



from .logger import EventLogger
from .metadata import get_session_id
from .publisher import GlobalAnalyticsBus

name = "analytics"


def _create_global_analytics_bus():
    return GlobalAnalyticsBus()


log = EventLogger(handler=_create_global_analytics_bus(), session_id=get_session_id())



import logging
from typing import Optional

from localstack.services.dynamodb import dynamodb_listener
from localstack.services.dynamodb.server import DynamodbServer, create_dynamodb_server
from localstack.services.infra import log_startup_message, start_proxy_for_service
from localstack.utils.aws import aws_stack

LOG = logging.getLogger(__name__)

# server singleton
_server: Optional[DynamodbServer] = None


def check_dynamodb(expect_shutdown=False, print_error=False):
    out = None

    if not expect_shutdown:
        assert _server

    try:
        _server.wait_is_up()
        out = aws_stack.connect_to_service("dynamodb", endpoint_url=_server.url).list_tables()
    except Exception:
        if print_error:
            LOG.exception("DynamoDB health check failed")
    if expect_shutdown:
        assert out is None
    else:
        assert isinstance(out["TableNames"], list)


def start_dynamodb(port=None, asynchronous=True, update_listener=None):
    global _server
    if not _server:
        _server = create_dynamodb_server()

    _server.start()

    log_startup_message("DynamoDB")
    start_proxy_for_service(
        "dynamodb",
        port,
        backend_port=_server.port,
        update_listener=update_listener,
    )
    return _server


def restart_dynamodb():
    global _server
    if _server:
        _server.shutdown()
        _server.join(timeout=10)
        _server = None

    LOG.debug("Restarting DynamoDB process ...")
    start_dynamodb(asynchronous=True, update_listener=dynamodb_listener.UPDATE_DYNAMODB)



import os
import threading
from typing import Callable, Optional

from .platform import is_windows
from .run import ShellCommandThread, is_command_available
from .threads import FuncThread


class FileListener:
    """
    Platform independent `tail -f` command that calls a callback every time a new line is received on the file. If
    use_tail_command is set (which is the default if we're not on windows and the tail command is available),
    then a `tail -f` subprocess will be started. Otherwise the tailer library is used that uses polling with retry.
    """

    def __init__(self, file_path: str, callback: Callable[[str], None]):
        self.file_path = file_path
        self.callback = callback

        self.thread: Optional[FuncThread] = None
        self.started = threading.Event()

        self.use_tail_command = not is_windows() and is_command_available("tail")

    def start(self):
        self.thread = self._do_start_thread()
        self.started.wait()

        if self.thread.result_future.done():
            # this will re-raise exceptions from the run command that occurred before started was set
            self.thread.result_future.result()

    def join(self, timeout=None):
        if self.thread:
            self.thread.join(timeout=timeout)

    def close(self):
        if self.thread and self.thread.running:
            self.thread.stop()

        self.started.clear()
        self.thread = None

    def _do_start_thread(self) -> FuncThread:
        if self.use_tail_command:
            thread = self._create_tail_command_thread()
            thread.start()
            thread.started.wait(5)
            self.started.set()
        else:
            thread = self._create_tailer_thread()
            thread.start()

        return thread

    def _create_tail_command_thread(self) -> ShellCommandThread:
        def _log_listener(line, *args, **kwargs):
            try:
                self.callback(line.rstrip("\r\n"))
            except Exception:
                pass

        if not os.path.isfile(self.file_path):
            raise FileNotFoundError

        return ShellCommandThread(
            cmd=["tail", "-f", self.file_path], quiet=False, log_listener=_log_listener
        )

    def _create_tailer_thread(self) -> FuncThread:
        from tailer import Tailer

        tailer = Tailer(open(self.file_path), end=True)

        def _run_follow(*_):
            try:
                self.started.set()
                for line in tailer.follow(delay=0.25):
                    try:
                        self.callback(line)
                    except Exception:
                        pass
            finally:
                tailer.close()

        return FuncThread(func=_run_follow, on_stop=lambda *_: tailer.close())



from localstack import config
from localstack.services.cloudformation import cloudformation_api
from localstack.services.infra import start_local_api


def start_cloudformation(port=None, asynchronous=False):
    port = port or config.service_port("cloudformation")
    return start_local_api(
        "CloudFormation",
        port,
        api="cloudformation",
        method=cloudformation_api.serve,
        asynchronous=asynchronous,
    )
